# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
========================================
Solvers (:mod:`qiskit_dynamics.solvers`)
========================================

.. currentmodule:: qiskit_dynamics.solvers

This module provides classes and functions for solving differential equations.

Table :ref:`1 <standard solver table>` summarizes the standard solver interfaces exposed in this
module. It includes a high level class :class:`~qiskit_dynamics.solvers.Solver` for solving models
of quantum systems, as well as low-level functions for solving both ordinary differential equations
:math:`\dot{y}(t) = f(t, y(t))` and linear matrix differential equations
:math:`\dot{y}(t) = G(t)y(t)`.

Additionally, this module contains more specialized solvers for linear matrix differential equations
based on perturbative expansions, described :ref:`below <perturbative solvers>`.

.. _standard solver table:

.. list-table:: Solver interfaces
   :widths: 10 50
   :header-rows: 1

   * - Object
     - Description
   * - :class:`~qiskit_dynamics.solvers.Solver`
     - High level solver class for both Hamiltonian and Lindblad dynamics. Automatically constructs
       the relevant model type based on system details, and the
       :meth:`~qiskit_dynamics.solvers.Solver.solve` method automatically handles
       ``qiskit.quantum_info`` input types.
   * - :func:`~qiskit_dynamics.solvers.solve_ode`
     - Low level solver function for ordinary differential equations:

       .. math::

            \dot{y}(t) = f(t, y(t)),

       for :math:`y(t)` arrays of arbitrary shape and :math:`f` specified as an arbitrary callable.
   * - :func:`~qiskit_dynamics.solvers.solve_lmde`
     - Low level solver function for linear matrix differential equations in *standard form*:

       .. math::
            \dot{y}(t) = G(t)y(t),

       where :math:`G(t)` is either a callable or a ``qiskit_dynamics`` model type, and :math:`y(t)`
       arrays of suitable shape for the matrix multiplication above.

.. _perturbative solvers:

Perturbative Solvers
====================

The classes :class:`~qiskit_dynamics.solvers.DysonSolver` and
:class:`~qiskit_dynamics.solvers.MagnusSolver` implement advanced solvers detailed in
:footcite:`puzzuoli_algorithms_2023`, with the :class:`~qiskit_dynamics.solvers.DysonSolver`
implementing a variant of the *Dysolve* algorithm originally introduced in
:footcite:p:`shillito_fast_2020`.

The solvers are specialized to linear matrix differential equations with :math:`G(t)`
decomposed as:

.. math::

    G(t) = G_0 + \sum_j Re[f_j(t)e^{i2\pi\nu_jt}]G_j,

and are fixed step with a pre-defined step size :math:`\Delta t`. The differential equation is
solved by either computing a truncated Dyson series, or taking the exponential of a truncated Magnus
expansion.

Add reference to both userguide and perturbation theory module documentation.


Solver classes
==============

.. autosummary::
   :toctree: ../stubs/

   Solver
   DysonSolver
   MagnusSolver

Solver functions
================

.. autosummary::
   :toctree: ../stubs/

   solve_ode
   solve_lmde

.. footbibliography::
"""

from .solver_functions import solve_ode, solve_lmde
from .solver_classes import Solver
from .perturbative_solvers.dyson_solver import DysonSolver
from .perturbative_solvers.magnus_solver import MagnusSolver



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Generator models module.
"""

from abc import ABC, abstractmethod
from typing import Union, List, Optional
from warnings import warn
from scipy.sparse import diags

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics.arraylias.alias import ArrayLike
from qiskit_dynamics.models.operator_collections import (
    OperatorCollection,
    ScipySparseOperatorCollection,
)
from qiskit_dynamics.signals import Signal, SignalList
from .rotating_frame import RotatingFrame

try:
    import jax
except ImportError:
    pass


class BaseGeneratorModel(ABC):
    r"""Defines an interface for a time-dependent linear differential equation of the form
    :math:`\dot{y}(t) = \Lambda(t, y)`, where :math:`\Lambda` is linear in :math:`y`. The core
    functionality is evaluation of :math:`\Lambda(t, y)`, as well as, if possible, evaluation of the
    map :math:`\Lambda(t, \cdot)`.
    """

    def __init__(self, array_library: Optional[str] = None):
        """Set up general information used by all subclasses."""
        self._array_library = array_library

    @property
    @abstractmethod
    def dim(self) -> int:
        """The matrix dimension."""

    @property
    @abstractmethod
    def rotating_frame(self) -> RotatingFrame:
        """The rotating frame."""

    @property
    @abstractmethod
    def in_frame_basis(self) -> bool:
        """Whether or not the model is evaluated in the basis in which the frame is diagonalized."""

    @property
    def array_library(self) -> Union[None, str]:
        """Array library with which to represent the operators in the model, and to evaluate the
        model.

        See the list of supported array libraries in the :mod:`.arraylias` submodule API
        documentation.
        """
        return self._array_library

    @abstractmethod
    def evaluate(self, time: float) -> ArrayLike:
        r"""If possible, evaluate the map :math:`\Lambda(t, \cdot)`.

        Args:
            time: The time to evaluate the model at.
        """

    @abstractmethod
    def evaluate_rhs(self, time: float, y: ArrayLike) -> ArrayLike:
        r"""Evaluate the right hand side :math:`\dot{y}(t) = \Lambda(t, y)`.

        Args:
            time: The time to evaluate the model at.
            y: State of the differential equation.
        """

    def __call__(self, time: float, y: Optional[ArrayLike] = None) -> ArrayLike:
        r"""Evaluate generator RHS functions. If ``y is None``, attemps to evaluate
        :math:`\Lambda(t, \cdot)`, otherwise, calculates :math:`\Lambda(t, y)`.

        Args:
            time: The time to evaluate at.
            y: Optional state.

        Returns:
            ArrayLike: Either the evaluated model, or the RHS for the given y.
        """
        return self.evaluate(time) if y is None else self.evaluate_rhs(time, y)


class GeneratorModel(BaseGeneratorModel):
    r"""A model for a a linear matrix differential equation in standard form.

    :class:`GeneratorModel` is a concrete instance of :class:`BaseGeneratorModel`, where the map
    :math:`\Lambda(t, y)` is explicitly constructed as:

    .. math::

        \Lambda(t, y) = G(t)y,

        G(t) = \sum_i s_i(t) G_i + G_d

    where the :math:`G_i` are matrices (represented by an :class:`Operator` or array),
    the :math:`s_i(t)` are signals represented by a list of :class:`Signal` objects, and :math:`G_d`
    is the constant-in-time static term of the generator.
    """

    def __init__(
        self,
        static_operator: Optional[ArrayLike] = None,
        operators: Optional[ArrayLike] = None,
        signals: Optional[Union[SignalList, List[Signal]]] = None,
        rotating_frame: Optional[Union[ArrayLike, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        array_library: Optional[str] = None,
    ):
        """Initialize.

        Args:
            static_operator: Constant part of the generator.
            operators: A list of operators :math:`G_i`.
            signals: Stores the terms :math:`s_i(t)`. While required for evaluation,
                :class:`GeneratorModel` signals are not required at instantiation.
            rotating_frame: Rotating frame operator.
            in_frame_basis: Whether to represent the model in the basis in which the rotating frame
                operator is diagonalized.
            array_library: Array library with which to represent the operators in the model, and to
                evaluate the model. See the list of supported array libraries in the
                :mod:`.arraylias` submodule API documentation. If ``None``, the arrays will be
                handled by general dispatching rules.
        Raises:
            QiskitError: If model not sufficiently specified.
        """
        if static_operator is None and operators is None:
            raise QiskitError(
                f"{type(self).__name__} requires at least one of static_operator or operators to "
                "be specified at construction."
            )

        self._rotating_frame = RotatingFrame(rotating_frame)
        self._in_frame_basis = in_frame_basis

        # set up internal operators
        static_operator = _static_operator_into_frame_basis(
            static_operator=static_operator,
            rotating_frame=self._rotating_frame,
            array_library=array_library,
        )

        operators = _operators_into_frame_basis(
            operators=operators, rotating_frame=self._rotating_frame, array_library=array_library
        )

        self._operator_collection = _get_operator_collection(
            static_operator=static_operator, operators=operators, array_library=array_library
        )

        self._signals = None
        self.signals = signals

        super().__init__(array_library=array_library)

    @property
    def dim(self) -> int:
        """The matrix dimension."""
        return self._operator_collection.dim

    @property
    def rotating_frame(self) -> RotatingFrame:
        """The rotating frame."""
        return self._rotating_frame

    @property
    def in_frame_basis(self) -> bool:
        """Whether or not the model is evaluated in the basis in which the frame is diagonalized."""
        return self._in_frame_basis

    @in_frame_basis.setter
    def in_frame_basis(self, in_frame_basis: bool):
        self._in_frame_basis = in_frame_basis

    @property
    def static_operator(self) -> Union[ArrayLike, None]:
        """The static operator."""
        if self._operator_collection.static_operator is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.static_operator
        return self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.static_operator
        )

    @property
    def operators(self) -> Union[ArrayLike, None]:
        """The operators in the model."""
        if self._operator_collection.operators is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.operators
        return self.rotating_frame.operator_out_of_frame_basis(self._operator_collection.operators)

    @property
    def signals(self) -> SignalList:
        """The signals in the model.

        Raises:
            QiskitError: If set to ``None`` when operators exist, or when set to a number of signals
                different then the number of operators.
        """
        return self._signals

    @signals.setter
    def signals(self, signals: Union[SignalList, List[Signal]]):
        if signals is None:
            self._signals = None
        elif signals is not None and self.operators is None:
            raise QiskitError("Signals must be None if operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(signals, list):
                signals = SignalList(signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(signals, SignalList):
                raise QiskitError("Signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.operators, list):
                len_operators = len(self.operators)
            else:
                len_operators = self.operators.shape[0]
            if len(signals) != len_operators:
                raise QiskitError("Signals needs to have the same length as operators.")

            self._signals = signals

    def evaluate(self, time: float) -> ArrayLike:
        """Evaluate the model in array format as a matrix, independent of state.

        Args:
            time: The time to evaluate the model at.

        Returns:
            ArrayLike: The evaluated model as a matrix.

        Raises:
            QiskitError: If model cannot be evaluated.
        """
        if self._signals is None and self._operator_collection.operators is not None:
            raise QiskitError(
                f"{type(self).__name__} with non-empty operators must be evaluated signals."
            )

        # Evaluated in frame basis, but without rotations
        op_combo = self._operator_collection(self._signals(time) if self._signals else None)

        # Apply rotations e^{-Ft}Ae^{Ft} in frame basis where F = D
        return self.rotating_frame.operator_into_frame(
            time, op_combo, operator_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
        )

    def evaluate_rhs(self, time: float, y: ArrayLike) -> ArrayLike:
        r"""Evaluate ``G(t) @ y``.

        Args:
            time: The time to evaluate the model at .
            y: Array specifying system state.

        Returns:
            Array defined by :math:`G(t) \times y`.

        Raises:
            QiskitError: If model cannot be evaluated.
        """
        if self._signals is None:
            if self._operator_collection.operators is not None:
                raise QiskitError(
                    f"{type(self).__name__} with non-empty operators must be evaluated signals."
                )
            sig_vals = None
        else:
            sig_vals = self._signals(time)

        if self.rotating_frame is not None:
            # First, compute e^{tF}y as a pre-rotation in the frame basis
            out = self.rotating_frame.state_out_of_frame(
                time, y, y_in_frame_basis=self._in_frame_basis, return_in_frame_basis=True
            )
            # Then, compute the product Ae^{tF}y
            out = self._operator_collection(sig_vals, out)
            # Finally, we have the full operator e^{-tF}Ae^{tF}y
            out = self.rotating_frame.state_into_frame(
                time, out, y_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
            )
            return out

        return self._operator_collection(sig_vals, y)


def _static_operator_into_frame_basis(
    static_operator: Union[None, ArrayLike],
    rotating_frame: RotatingFrame,
    array_library: Optional[ArrayLike] = None,
) -> Union[None, ArrayLike]:
    """Converts the static_operator into the frame basis, including a subtraction of the frame
    operator. This function also enforces typing via array_library.
    """

    # handle static_operator is None case
    if static_operator is None:
        if rotating_frame.frame_operator is None:
            return None
        if array_library == "scipy_sparse":
            return -diags(rotating_frame.frame_diag, format="csr")
        return unp.diag(-rotating_frame.frame_diag)

    static_operator = numpy_alias(like=array_library).asarray(static_operator)

    return rotating_frame.generator_into_frame(
        t=0.0, operator=static_operator, return_in_frame_basis=True
    )


def _operators_into_frame_basis(
    operators: Union[None, list, ArrayLike],
    rotating_frame: RotatingFrame,
    array_library: Optional[str] = None,
) -> Union[None, ArrayLike]:
    """Converts operators into the frame basis. This function also enforces typing via
    array_library.
    """
    if operators is None:
        return None

    if array_library == "scipy_sparse" or (
        array_library is None and "scipy_sparse" in numpy_alias.infer_libs(operators)
    ):
        ops = []
        for op in operators:
            op = numpy_alias(like="scipy_sparse").asarray(op)
            ops.append(rotating_frame.operator_into_frame_basis(op))
        return ops

    return rotating_frame.operator_into_frame_basis(
        numpy_alias(like=array_library).asarray(operators)
    )


def _get_operator_collection(
    static_operator: Union[None, ArrayLike],
    operators: Union[None, ArrayLike],
    array_library: Optional[str] = None,
) -> Union[OperatorCollection, ScipySparseOperatorCollection]:
    """Construct an operator collection for :class:`GeneratorModel`.

    Args:
        static_operator: Static operator of the model.
        operators: Operators for the model.
        array_library: Array library to use.

    Returns:
        Union[OperatorCollection, ScipySparseOperatorCollection]: The relevant operator collection.
    """

    if array_library == "scipy_sparse":
        return ScipySparseOperatorCollection(static_operator=static_operator, operators=operators)

    if array_library == "jax_sparse":
        # warn that sparse mode when using JAX is primarily recommended for use on CPU
        if jax.default_backend() != "cpu":
            warn(
                """Using sparse mode with JAX is primarily recommended for use on CPU.""",
                stacklevel=2,
            )

    return OperatorCollection(
        static_operator=static_operator, operators=operators, array_library=array_library
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""Operator collections as math/calculation objects for model classes."""

from typing import Any, Union, List, Optional
import numpy as np
from scipy.sparse import csr_matrix

from qiskit import QiskitError
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics.arraylias.alias import ArrayLike, _numpy_multi_dispatch
from .model_utils import vec_commutator, vec_dissipator


def _linear_combo(coeffs, mats):
    return _numpy_multi_dispatch(coeffs, mats, path="linear_combo")


def _matmul(A, B, **kwargs):
    return _numpy_multi_dispatch(A, B, path="matmul", **kwargs)


def _to_csr_object_array(ops, decimals):
    """Turn a list of matrices into a numpy object array of scipy sparse matrix instances."""
    if ops is None:
        return None
    return np.array(
        [numpy_alias(like="scipy_sparse").asarray(np.round(op, decimals)) for op in ops]
    )


class OperatorCollection:
    r"""Class for evaluating a linear combination of operators acting on a state.

    This class represents a function :math:`c,y \mapsto \Lambda(c, y)`, which is assumed to be
    decomposed as :math:`\Lambda(c, y) = (G_d + \sum_jc_jG_j)  y` for matrices :math:`G_d` and
    :math:`G_j`, with :math:`G_d` referred to as the static operator.

    This works for ``array_library in ["numpy", "jax", "jax_sparse"]``.
    """

    def __init__(
        self,
        static_operator: Optional[ArrayLike] = None,
        operators: Optional[ArrayLike] = None,
        array_library: Optional[str] = None,
    ):
        """Initialize.

        Args:
            operators: ``(k,n,n)`` array specifying the terms :math:`G_j`.
            static_operator: ``(n,n)`` array specifying the extra static_operator :math:`G_d`.
            array_library: Underlying library to use for array operations.

        Raises:
            QiskitError: If "scipy_sparse" is passed as array_library.
        """
        if array_library == "scipy_sparse":
            raise QiskitError("scipy_sparse is not a valid array_library for OperatorCollection.")

        if static_operator is not None:
            self._static_operator = numpy_alias(like=array_library).asarray(static_operator)
        else:
            self._static_operator = None

        if operators is not None:
            self._operators = numpy_alias(like=array_library).asarray(operators)
        else:
            self._operators = None

    @property
    def dim(self) -> int:
        """The matrix dimension."""
        if self.static_operator is not None:
            return self.static_operator.shape[-1]
        else:
            return self.operators[0].shape[-1]

    @property
    def static_operator(self) -> Union[ArrayLike, None]:
        """The static part of the operator collection."""
        return self._static_operator

    @property
    def operators(self) -> Union[ArrayLike, None]:
        """The operators of this collection."""
        return self._operators

    def evaluate(self, coefficients: Union[ArrayLike, None]) -> ArrayLike:
        r"""Evaluate the operator :math:`\Lambda(c, \cdot) = (G_d + \sum_jc_jG_j)`.

        Args:
            coefficients: The coefficient values :math:`c`.

        Returns:
            An ``ArrayLike`` that acts on states ``y`` via multiplication.

        Raises:
            QiskitError: If both static_operator and operators are ``None``.
        """
        if self._static_operator is not None and self._operators is not None:
            return _linear_combo(coefficients, self._operators) + self._static_operator
        elif self._static_operator is None and self._operators is not None:
            return _linear_combo(coefficients, self._operators)
        elif self._static_operator is not None:
            return self._static_operator
        raise QiskitError(
            "OperatorCollection with None for both static_operator and operators cannot be "
            "evaluated."
        )

    def evaluate_rhs(self, coefficients: Union[ArrayLike, None], y: ArrayLike) -> ArrayLike:
        r"""Evaluate the function and return :math:`\Lambda(c, y) = (G_d + \sum_jc_jG_j)  y`.

        Args:
            coefficients: The coefficients :math:`c`.
            y: The system state.

        Returns:
            The evaluated function.
        """
        return _matmul(self.evaluate(coefficients), y)

    def __call__(
        self, coefficients: Union[ArrayLike, None], y: Optional[ArrayLike] = None
    ) -> ArrayLike:
        """Call :meth:`~evaluate` or :meth:`~evaluate_rhs` depending on the presense of ``y``.

        Args:
            coefficients: The coefficients :math:`c` to use for evaluation.
            y: Optionally, the system state.

        Returns:
            The output of :meth:`~evaluate` or :meth:`self.evaluate_rhs`.
        """
        return self.evaluate(coefficients) if y is None else self.evaluate_rhs(coefficients, y)


class ScipySparseOperatorCollection:
    r"""Scipy sparse version of :class:`OperatorCollection`."""

    def __init__(
        self,
        static_operator: Optional[ArrayLike] = None,
        operators: Optional[ArrayLike] = None,
        decimals: Optional[int] = 10,
    ):
        """Initialize.

        Args:
            static_operator: (n,n) Array specifying the static_operator term :math:`G_d`.
            operators: (k,n,n) Array specifying the terms :math:`G_j`.
            decimals: Values will be rounded at ``decimals`` places after decimal.
        """
        if static_operator is not None:
            self._static_operator = numpy_alias(like="scipy_sparse").asarray(
                np.round(static_operator, decimals)
            )
        else:
            self._static_operator = None

        self._operators = _to_csr_object_array(operators, decimals)

    @property
    def dim(self) -> int:
        """The matrix dimension."""
        if self.static_operator is not None:
            return self.static_operator.shape[-1]
        else:
            return self.operators[0].shape[-1]

    @property
    def static_operator(self) -> Union[None, csr_matrix]:
        """The static part of the operator collection."""
        return self._static_operator

    @property
    def operators(self) -> Union[None, List[csr_matrix]]:
        """The operators of this collection."""
        if self._operators is None:
            return None

        return list(self._operators)

    def evaluate(self, coefficients: Union[ArrayLike, None]) -> csr_matrix:
        r"""Evaluate the operator :math:`\Lambda(c, \cdot) = (G_d + \sum_jc_jG_j)`.

        Args:
            coefficients: The coefficient values :math:`c` to use on the operators.

        Returns:
            A scipy sparse matrix that acts on states ``y`` via multiplication.

        Raises:
            QiskitError: If collection cannot be evaluated.
        """
        if self._static_operator is not None and self._operators is not None:
            return (
                np.tensordot(coefficients, self._operators, axes=1).item() + self._static_operator
            )
        elif self._static_operator is None and self._operators is not None:
            return np.tensordot(coefficients, self._operators, axes=1).item()
        elif self.static_operator is not None:
            return self._static_operator
        raise QiskitError(
            f"{type(self).__name__} with None for both static_operator and operators cannot be "
            "evaluated."
        )

    def evaluate_rhs(self, coefficients: Union[ArrayLike, None], y: ArrayLike) -> ArrayLike:
        r"""Evaluate the function and return :math:`\Lambda(c, y) = (G_d + \sum_jc_jG_j)  y`.

        Args:
            coefficients: The coefficients :math:`c` to use on the operators.
            y: The system state.

        Returns:
            The evaluated function.
        Raises:
            QiskitError: If the function cannot be evaluated.
        """
        if len(y.shape) == 2:
            # For 2d array, compute linear combination then multiply
            gen = self.evaluate(coefficients)
            return gen.dot(y)
        elif len(y.shape) == 1:
            # For a 1d array, multiply individual matrices then compute linear combination
            tmparr = np.empty(shape=(1), dtype="O")
            tmparr[0] = y

            if self._static_operator is not None and self._operators is not None:
                return np.dot(coefficients, self._operators * tmparr) + self.static_operator.dot(y)
            elif self._static_operator is None and self._operators is not None:
                return np.dot(coefficients, self._operators * tmparr)
            elif self.static_operator is not None:
                return self.static_operator.dot(y)

            raise QiskitError(
                self.__class__.__name__
                + """  with None for both static_operator and
                                operators cannot be evaluated."""
            )

        raise QiskitError(self.__class__.__name__ + """  cannot evaluate RHS for y.ndim > 3.""")

    def __call__(
        self, coefficients: Union[ArrayLike, None], y: Optional[ArrayLike] = None
    ) -> ArrayLike:
        """Call :meth:`~evaluate` or :meth:`~evaluate_rhs` depending on the presense of ``y``.

        Args:
            coefficients: The coefficients :math:`c` to use for evaluation.
            y: Optionally, the system state.

        Returns:
            The output of :meth:`~evaluate` or :meth:`self.evaluate_rhs`.
        """
        return self.evaluate(coefficients) if y is None else self.evaluate_rhs(coefficients, y)


class LindbladCollection:
    r"""Class representing a two-variable matrix function for evaluating the right hand
    side of the Lindblad equation.

    In particular, this object represents the function:
    .. math::
        \Lambda(c_1, c_2, \rho) = -i[H_d + \sum_j c_{1,j}H_j,\rho]
                                    + \sum_j(D_j\rho D_j^\dagger
                                        - (1/2) * {D_j^\daggerD_j,\rho})
                                    + \sum_jc_{2,j}(L_j\rho L_j^\dagger
                                    - (1/2) * {L_j^\daggerL_j,\rho})

    where :math:`\[,\]` and :math:`\{,\}` are the operator
    commutator and anticommutator, respectively.

    Describes an interface for evaluating the map or its action on :math:`\rho`,
    given a pair of 1d sets of values :math:`c_1, c_2`.

    This class works for ``array_library in ["numpy", "jax", "jax_sparse"]``.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        array_library: Optional[str] = None,
    ):
        r"""Initialize collection. Argument types depend on concrete subclass.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            static_dissipators: Constant dissipator terms.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
            array_library: Array library to use for storing arrays in the collection.

        Raises:
            QiskitError: If "scipy_sparse" is passed as the array_library.
        """

        if array_library == "scipy_sparse":
            raise QiskitError("scipy_sparse is not a valid array_library for LindbladCollection.")

        if static_hamiltonian is not None:
            self._static_hamiltonian = numpy_alias(like=array_library).asarray(static_hamiltonian)
        else:
            self._static_hamiltonian = None

        if hamiltonian_operators is not None:
            self._hamiltonian_operators = numpy_alias(like=array_library).asarray(
                hamiltonian_operators
            )
        else:
            self._hamiltonian_operators = None

        if static_dissipators is not None:
            if array_library == "jax_sparse":
                self._static_dissipators = numpy_alias(like="jax").asarray(static_dissipators)
            else:
                self._static_dissipators = numpy_alias(like=array_library).asarray(
                    static_dissipators
                )

            self._static_dissipators_adj = unp.conjugate(
                unp.transpose(self._static_dissipators, [0, 2, 1])
            )
            self._static_dissipators_product_sum = -0.5 * unp.sum(
                unp.matmul(self._static_dissipators_adj, self._static_dissipators), axis=0
            )

            if array_library == "jax_sparse":
                from jax.experimental.sparse import BCOO

                self._static_dissipators = BCOO.fromdense(self._static_dissipators, n_batch=1)
                self._static_dissipators_adj = BCOO.fromdense(
                    self._static_dissipators_adj, n_batch=1
                )
                self._static_dissipators_product_sum = BCOO.fromdense(
                    self._static_dissipators_product_sum
                )
        else:
            self._static_dissipators = None

        if dissipator_operators is not None:
            if array_library == "jax_sparse":
                self._dissipator_operators = numpy_alias(like="jax").asarray(dissipator_operators)
            else:
                self._dissipator_operators = numpy_alias(like=array_library).asarray(
                    dissipator_operators
                )

            self._dissipator_operators_adj = unp.conjugate(
                unp.transpose(self._dissipator_operators, [0, 2, 1])
            )
            self._dissipator_products = -0.5 * unp.matmul(
                self._dissipator_operators_adj, self._dissipator_operators
            )

            if array_library == "jax_sparse":
                from jax.experimental.sparse import BCOO

                self._dissipator_operators = BCOO.fromdense(self._dissipator_operators, n_batch=1)
                self._dissipator_operators_adj = BCOO.fromdense(
                    self._dissipator_operators_adj, n_batch=1
                )
                self._dissipator_products = BCOO.fromdense(self._dissipator_products, n_batch=1)
        else:
            self._dissipator_operators = None

    @property
    def static_hamiltonian(self) -> ArrayLike:
        """The static part of the Hamiltonian."""
        return self._static_hamiltonian

    @property
    def hamiltonian_operators(self) -> ArrayLike:
        """The operators for the non-static part of Hamiltonian."""
        return self._hamiltonian_operators

    @property
    def static_dissipators(self) -> ArrayLike:
        """The operators for the static part of dissipator."""
        return self._static_dissipators

    @property
    def dissipator_operators(self) -> ArrayLike:
        """The operators for the non-static part of dissipator."""
        return self._dissipator_operators

    def evaluate_hamiltonian(self, ham_coefficients: Optional[ArrayLike]) -> ArrayLike:
        r"""Evaluate the Hamiltonian of the model.

        Args:
            ham_coefficients: The values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.

        Returns:
            The Hamiltonian.

        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                _linear_combo(ham_coefficients, self._hamiltonian_operators)
                + self._static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return _linear_combo(ham_coefficients, self._hamiltonian_operators)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    def evaluate(
        self, ham_coefficients: Optional[ArrayLike], dis_coefficients: Optional[ArrayLike]
    ) -> ArrayLike:
        r"""Evaluate the function and return :math:`\Lambda(c_1, c_2, \cdot)`.

        Args:
            ham_coefficients: The signals :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The signals :math:`c_2` to use on the dissipators.

        Returns:
            The evaluated function.

        Raises:
            ValueError: Always.
        """
        raise ValueError("Non-vectorized Lindblad collections cannot be evaluated without a state.")

    def evaluate_rhs(
        self,
        ham_coefficients: Optional[ArrayLike],
        dis_coefficients: Optional[ArrayLike],
        y: ArrayLike,
    ) -> ArrayLike:
        r"""Evaluates Lindblad equation RHS given a pair of signal values for the hamiltonian terms
        and the dissipator terms.

        Expresses the RHS of the Lindblad equation as :math:`(A+B)y + y(A-B) + C`, where

        .. math::
            A = (-1/2)*\sum_jD_j^\dagger D_j + (-1/2)*\sum_j\gamma_j(t) L_j^\dagger L_j,

            B = -iH,

            C = \sum_j \gamma_j(t) L_j y L_j^\dagger.

        Args:
            ham_coefficients: Hamiltonian coefficient values, :math:`s_j(t)`.
            dis_coefficients: Dissipator coefficient values, :math:`\gamma_j(t)`.
            y: Density matrix as ``(n,n)`` array representing the state at time :math:`t`.

        Returns:
            RHS of the Lindblad equation
            .. math::
                -i[H,y] + \sum_j\gamma_j(t)(L_j y L_j^\dagger - (1/2) * {L_j^\daggerL_j,y}).

        Raises:
            QiskitError: If operator collection is underspecified.
        """

        hamiltonian_matrix = None
        if self._static_hamiltonian is not None or self._hamiltonian_operators is not None:
            hamiltonian_matrix = -1j * self.evaluate_hamiltonian(ham_coefficients)  # B matrix

        # if dissipators present (includes both hamiltonian is None and is not None)
        if self._dissipator_operators is not None or self._static_dissipators is not None:
            # A matrix
            if self._static_dissipators is None:
                dissipators_matrix = _linear_combo(dis_coefficients, self._dissipator_products)
            elif self._dissipator_operators is None:
                dissipators_matrix = self._static_dissipators_product_sum
            else:
                dissipators_matrix = self._static_dissipators_product_sum + _linear_combo(
                    dis_coefficients, self._dissipator_products
                )

            if hamiltonian_matrix is not None:
                left_mult_contribution = _matmul(hamiltonian_matrix + dissipators_matrix, y)
                right_mult_contribution = _matmul(y, dissipators_matrix - hamiltonian_matrix)
            else:
                left_mult_contribution = _matmul(dissipators_matrix, y)
                right_mult_contribution = _matmul(y, dissipators_matrix)

            if len(y.shape) == 3:
                # Must do array broadcasting and transposition to ensure vectorization works
                y = unp.broadcast_to(y, (1, y.shape[0], y.shape[1], y.shape[2])).transpose(
                    [1, 0, 2, 3]
                )

            if self._static_dissipators is None:
                # real and imag broken up to avoid real/complex tensordot warning
                mats = _matmul(
                    self._dissipator_operators, _matmul(y, self._dissipator_operators_adj)
                )
                both_mult_contribution = _numpy_multi_dispatch(
                    dis_coefficients,
                    mats.real,
                    path="tensordot",
                    axes=(-1, -3),
                ) + 1j * _numpy_multi_dispatch(
                    dis_coefficients,
                    mats.imag,
                    path="tensordot",
                    axes=(-1, -3),
                )
            elif self._dissipator_operators is None:
                both_mult_contribution = unp.sum(
                    _matmul(self._static_dissipators, _matmul(y, self._static_dissipators_adj)),
                    axis=-3,
                )
            else:
                # real and imag broken up to avoid real/complex tensordot warning
                mats = _matmul(
                    self._dissipator_operators, _matmul(y, self._dissipator_operators_adj)
                )
                both_mult_contribution = (
                    unp.sum(
                        _matmul(self._static_dissipators, _matmul(y, self._static_dissipators_adj)),
                        axis=-3,
                    )
                    + _numpy_multi_dispatch(
                        dis_coefficients,
                        mats.real,
                        path="tensordot",
                        axes=(-1, -3),
                    )
                    + 1j
                    * _numpy_multi_dispatch(
                        dis_coefficients,
                        mats.imag,
                        path="tensordot",
                        axes=(-1, -3),
                    )
                )

            return left_mult_contribution + right_mult_contribution + both_mult_contribution
        # if just hamiltonian
        elif hamiltonian_matrix is not None:
            return _matmul(hamiltonian_matrix, y) - _matmul(y, hamiltonian_matrix)
        else:
            raise QiskitError(
                """LindbladCollection with None for static_hamiltonian,
                                 hamiltonian_operators, static_dissipators, and
                                 dissipator_operators, cannot evaluate rhs."""
            )

    def __call__(
        self,
        ham_coefficients: Optional[ArrayLike],
        dis_coefficients: Optional[ArrayLike],
        y: Optional[ArrayLike],
    ) -> ArrayLike:
        """Call :meth:`~evaluate` or :meth:`~evaluate_rhs` depending on the presense of ``y``.

        Args:
            ham_coefficients: The coefficients :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The coefficients :math:`c_2` to use on the dissipators.
            y: Optionally, the system state.

        Returns:
            The evaluated function.
        """
        if y is None:
            return self.evaluate(ham_coefficients, dis_coefficients)

        return self.evaluate_rhs(ham_coefficients, dis_coefficients, y)


class ScipySparseLindbladCollection:
    """Scipy sparse version of LindbladCollection."""

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        decimals: Optional[int] = 10,
    ):
        r"""Initializes sparse lindblad collection.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
            decimals: operator values will be rounded to ``decimals`` places after the
                decimal place to avoid excess storage of near-zero values
                in sparse format.
        """
        if static_hamiltonian is not None:
            self._static_hamiltonian = numpy_alias(like="scipy_sparse").asarray(
                np.round(static_hamiltonian, decimals)
            )
        else:
            self._static_hamiltonian = None

        self._hamiltonian_operators = _to_csr_object_array(hamiltonian_operators, decimals)
        self._static_dissipators = _to_csr_object_array(static_dissipators, decimals)
        self._dissipator_operators = _to_csr_object_array(dissipator_operators, decimals)

        if static_dissipators is not None:
            # setup adjoints
            self._static_dissipators_adj = np.array(
                [op.conj().transpose() for op in self._static_dissipators]
            )

            # pre-compute product and sum
            self._static_dissipators_product_sum = -0.5 * np.sum(
                self._static_dissipators_adj * self._static_dissipators, axis=0
            )

        if self._dissipator_operators is not None:
            # setup adjoints
            self._dissipator_operators_adj = np.array(
                [op.conj().transpose() for op in self._dissipator_operators]
            )

            # pre-compute products
            self._dissipator_products = (
                -0.5 * self._dissipator_operators_adj * self._dissipator_operators
            )

    @property
    def static_hamiltonian(self) -> Union[None, csr_matrix]:
        """The static part of the operator collection."""
        return self._static_hamiltonian

    @property
    def hamiltonian_operators(self) -> Union[None, list]:
        """The operators for the non-static part of Hamiltonian."""
        if self._hamiltonian_operators is None:
            return None

        return list(self._hamiltonian_operators)

    @property
    def static_dissipators(self) -> Union[None, list]:
        """The operators for the static part of dissipator."""
        if self._static_dissipators is None:
            return None

        return list(self._static_dissipators)

    @property
    def dissipator_operators(self) -> Union[None, list]:
        """The operators for the non-static part of dissipator."""
        if self._dissipator_operators is None:
            return None

        return list(self._dissipator_operators)

    def evaluate_hamiltonian(self, ham_coefficients: Optional[ArrayLike]) -> csr_matrix:
        r"""Compute the Hamiltonian.

        Args:
            ham_coefficients: The values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.
        Returns:
            Hamiltonian matrix.
        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                np.sum(ham_coefficients * self._hamiltonian_operators, axis=-1)
                + self.static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return np.sum(ham_coefficients * self._hamiltonian_operators, axis=-1)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    def evaluate(
        self, ham_coefficients: Optional[ArrayLike], dis_coefficients: Optional[ArrayLike]
    ) -> ArrayLike:
        r"""Evaluate the function and return :math:`\Lambda(c_1, c_2, \cdot)`.

        Args:
            ham_coefficients: The signals :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The signals :math:`c_2` to use on the dissipators.

        Returns:
            The evaluated function.

        Raises:
            ValueError: Always.
        """
        raise ValueError("Non-vectorized Lindblad collections cannot be evaluated without a state.")

    def evaluate_rhs(
        self,
        ham_coefficients: Optional[ArrayLike],
        dis_coefficients: Optional[ArrayLike],
        y: ArrayLike,
    ) -> ArrayLike:
        r"""Evaluate the RHS of the Lindblad model for a given list of signal values.

        Args:
            ham_coefficients: Stores Hamiltonian signal values :math:`s_j(t)`.
            dis_coefficients: Stores dissipator signal values :math:`\gamma_j(t)`. Pass ``None`` if
                no dissipator operators are involved.
            y: Density matrix of the system, a ``(k,n,n)`` Array.

        Returns:
            RHS of the Lindbladian.

        Raises:
            QiskitError: If RHS cannot be evaluated due to insufficient collection data.

        Calculation details:
            * for csr_matrices is equivalent to matrix multiplicaiton.
            We use numpy array broadcasting rules, combined with the above
            fact, to achieve speeds that are substantially faster than a for loop.
            First, in the case of a single (n,n) density matrix, we package the entire
            array as a single-element array whose entry is the array. In the case of
            multiple density matrices a (k,n,n) Array, we package everything as a
            (k,1) Array whose [j,0] entry is the [j,:,:] density matrix.

            In calculating the left- and right-mult contributions, we package
            H+L and H-L as (1) object arrays whose single entry stores the relevant
            sparse matrix. We can then multiply our packaged density matrix and
            [H\pm L]. Using numpy broadcasting rules, [H\pm L] will be broadcast
            to a (k,1) Array for elementwise multiplication with our packaged density
            matrices. After this, elementwise multiplication is applied. This in turn
            references each object's __mul__ function, which–for our csr_matrix components
            means matrix multiplication.

            In calculating the left-right-multiplication part, we use our (m)-shape
            object arrays holding the dissipator operators to perform multiplication.
            We can take an elementwise product with our packaged density matrix, at which
            point our dissipator operators are broadcast as (m) -> (1,m) -> (k,m) shaped,
            and our packaged density matrix as (k,1) -> (k,m). Elementwise multiplication
            is then applied, which is interpreted as matrix multiplication. This yields
            an array where entry [i,j] is an object storing the results of s_jL_j\rho_i L_j^\dagger.
            We can then sum over j and unpackage our object array to get our desired result.
        """
        hamiltonian_matrix = None
        if self._static_hamiltonian is not None or self._hamiltonian_operators is not None:
            hamiltonian_matrix = -1j * self.evaluate_hamiltonian(ham_coefficients)  # B matrix

        # package (n,n) Arrays as (1)
        # Arrays of dtype object, or (k,n,n) Arrays as (k,1) Arrays of dtype object
        y = package_density_matrices(y)

        # if dissipators present (includes both hamiltonian is None and is not None)
        if self._dissipator_operators is not None or self._static_dissipators is not None:
            # A matrix
            if self._static_dissipators is None:
                dissipators_matrix = np.sum(dis_coefficients * self._dissipator_products, axis=-1)
            elif self._dissipator_operators is None:
                dissipators_matrix = self._static_dissipators_product_sum
            else:
                dissipators_matrix = self._static_dissipators_product_sum + np.sum(
                    dis_coefficients * self._dissipator_products, axis=-1
                )

            if hamiltonian_matrix is not None:
                left_mult_contribution = np.squeeze([hamiltonian_matrix + dissipators_matrix] * y)
                right_mult_contribution = np.squeeze(y * [dissipators_matrix - hamiltonian_matrix])
            else:
                left_mult_contribution = np.squeeze([dissipators_matrix] * y)
                right_mult_contribution = np.squeeze(y * [dissipators_matrix])

            # both_mult_contribution[i] = \gamma_i L_i\rho L_i^\dagger performed in array language
            if self._static_dissipators is None:
                both_mult_contribution = np.sum(
                    (dis_coefficients * self._dissipator_operators)
                    * y
                    * self._dissipator_operators_adj,
                    axis=-1,
                )
            elif self._dissipator_operators is None:
                both_mult_contribution = np.sum(
                    self._static_dissipators * y * self._static_dissipators_adj, axis=-1
                )
            else:
                both_mult_contribution = np.sum(
                    (dis_coefficients * self._dissipator_operators)
                    * y
                    * self._dissipator_operators_adj,
                    axis=-1,
                ) + np.sum(self._static_dissipators * y * self._static_dissipators_adj, axis=-1)

            out = left_mult_contribution + right_mult_contribution + both_mult_contribution

        elif hamiltonian_matrix is not None:
            out = (([hamiltonian_matrix] * y) - (y * [hamiltonian_matrix]))[0]
        else:
            raise QiskitError(
                "ScipySparseLindbladCollection with None for static_hamiltonian, "
                "hamiltonian_operators, and dissipator_operators, cannot evaluate rhs."
            )
        if len(y.shape) == 2:
            # Very slow; avoid if not necessary (or if better implementation found). Needs to
            # map a (k) Array of dtype object with j^{th} entry a (n,n) Array -> (k,n,n) Array.
            out = unpackage_density_matrices(out.reshape(y.shape[0], 1))

        return out

    def __call__(
        self,
        ham_coefficients: Union[None, ArrayLike],
        dis_coefficients: Union[None, ArrayLike],
        y: Optional[ArrayLike],
    ) -> ArrayLike:
        """Call :meth:`~evaluate` or :meth:`~evaluate_rhs` depending on the presense of ``y``.

        Args:
            ham_coefficients: The signals :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The signals :math:`c_2` to use on the dissipators.
            y: Optionally, the system state.

        Returns:
            The evaluated function.
        """
        if y is None:
            return self.evaluate(ham_coefficients, dis_coefficients)

        return self.evaluate_rhs(ham_coefficients, dis_coefficients, y)


class VectorizedLindbladCollection:
    """Vectorized Lindblad collection class.

    The vectorized Lindblad equation represents the Lindblad master equation in the structure
    of a linear matrix differential equation in standard form.

    This class works for ``array_library in ["numpy", "jax", "jax_sparse"]``.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        array_library: Optional[str] = None,
    ):
        r"""Initialize collection.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            static_dissipators: Dissipator terms with coefficient 1.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.

        Raises:
            QiskitError: If "scipy_sparse" is passed as array_library.
        """

        self._array_library = array_library

        if array_library == "scipy_sparse":
            raise QiskitError("scipy_sparse is not a valid array_library for OperatorCollection.")

        if static_hamiltonian is not None:
            self._static_hamiltonian = self._convert_to_array_type(static_hamiltonian)
            self._vec_static_hamiltonian = vec_commutator(self._static_hamiltonian)
        else:
            self._static_hamiltonian = None

        if hamiltonian_operators is not None:
            self._hamiltonian_operators = self._convert_to_array_type_list(hamiltonian_operators)
            self._vec_hamiltonian_operators = vec_commutator(self._hamiltonian_operators)
        else:
            self._hamiltonian_operators = None

        if static_dissipators is not None:
            self._static_dissipators = self._convert_to_array_type_list(static_dissipators)
            self._vec_static_dissipators_sum = unp.sum(
                vec_dissipator(self._static_dissipators), axis=0
            )
        else:
            self._static_dissipators = None

        if dissipator_operators is not None:
            self._dissipator_operators = self._convert_to_array_type_list(dissipator_operators)
            self._vec_dissipator_operators = vec_dissipator(self._dissipator_operators)
        else:
            self._dissipator_operators = None

        # concatenate static operators
        if self._static_hamiltonian is not None and self._static_dissipators is not None:
            static_operator = self._vec_static_hamiltonian + self._vec_static_dissipators_sum
        elif self._static_hamiltonian is None and self._static_dissipators is not None:
            static_operator = self._vec_static_dissipators_sum
        elif self._static_hamiltonian is not None and self._static_dissipators is None:
            static_operator = self._vec_static_hamiltonian
        else:
            static_operator = None

        # concatenate non-static operators
        if self._hamiltonian_operators is not None and self._dissipator_operators is not None:
            operators = unp.append(
                self._vec_hamiltonian_operators, self._vec_dissipator_operators, axis=0
            )
        elif self._hamiltonian_operators is not None and self._dissipator_operators is None:
            operators = self._vec_hamiltonian_operators
        elif self._hamiltonian_operators is None and self._dissipator_operators is not None:
            operators = self._vec_dissipator_operators
        else:
            operators = None

        # build internally used operator collection
        self._operator_collection = self._construct_operator_collection(
            static_operator=static_operator, operators=operators
        )

    @property
    def static_hamiltonian(self) -> Union[ArrayLike, None]:
        """The static part of the operator collection."""
        return self._static_hamiltonian

    @property
    def hamiltonian_operators(self) -> Union[ArrayLike, None]:
        """The operators for the non-static part of Hamiltonian."""
        return self._hamiltonian_operators

    @property
    def static_dissipators(self) -> Union[ArrayLike, None]:
        """The operators for the static part of dissipator."""
        return self._static_dissipators

    @property
    def dissipator_operators(self) -> Union[ArrayLike, None]:
        """The operators for the non-static part of dissipator."""
        return self._dissipator_operators

    def evaluate_hamiltonian(self, ham_coefficients: Union[None, ArrayLike]) -> ArrayLike:
        r"""Evaluate the Hamiltonian of the model.

        Args:
            ham_coefficients: The values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.

        Returns:
            The Hamiltonian.

        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                _linear_combo(ham_coefficients, self._hamiltonian_operators)
                + self._static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return _linear_combo(ham_coefficients, self._hamiltonian_operators)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    def evaluate(
        self, ham_coefficients: Optional[ArrayLike], dis_coefficients: Optional[ArrayLike]
    ) -> ArrayLike:
        r"""Compute and return :math:`\Lambda(c_1, c_2, \cdot)`.

        Args:
            ham_coefficients: The signal coefficients :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The signal coefficients :math:`c_2` to use on the dissipators.

        Returns:
            The evaluated function.
        """
        coeffs = self._concatenate_coefficients(ham_coefficients, dis_coefficients)
        return self._operator_collection.evaluate(coeffs)

    def evaluate_rhs(
        self,
        ham_coefficients: Optional[ArrayLike],
        dis_coefficients: Optional[ArrayLike],
        y: ArrayLike,
    ) -> ArrayLike:
        r"""Evaluates the RHS of the Lindblad equation using vectorized maps.

        Args:
            ham_coefficients: Hamiltonian signal coefficients.
            dis_coefficients: Dissipator signal coefficients. If none involved, pass ``None``.
            y: Density matrix represented as a vector using column-stacking convention.

        Returns:
            Vectorized RHS of Lindblad equation :math:`\dot{\rho}` in column-stacking convention.
        """
        coeffs = self._concatenate_coefficients(ham_coefficients, dis_coefficients)
        return self._operator_collection.evaluate_rhs(coeffs, y)

    def __call__(
        self,
        ham_coefficients: Optional[ArrayLike],
        dis_coefficients: Optional[ArrayLike],
        y: Optional[ArrayLike],
    ) -> ArrayLike:
        """Call :meth:`~evaluate` or :meth:`~evaluate_rhs` depending on the presense of ``y``.

        Args:
            ham_coefficients: The signal coefficients :math:`c_1` to use on the Hamiltonians.
            dis_coefficients: The signal coefficients :math:`c_2` to use on the dissipators.
            y: Optionally, the system state.

        Returns:
            The evaluated function.
        """
        if y is None:
            return self.evaluate(ham_coefficients, dis_coefficients)

        return self.evaluate_rhs(ham_coefficients, dis_coefficients, y)

    def _convert_to_array_type(self, obj: Any) -> ArrayLike:
        return numpy_alias(like=self._array_library).asarray(obj)

    def _convert_to_array_type_list(self, obj: Any) -> ArrayLike:
        return numpy_alias(like=self._array_library).asarray(obj)

    def _construct_operator_collection(self, *args, **kwargs):
        """The class used for evaluating the vectorized model or RHS."""
        return OperatorCollection(*args, **kwargs, array_library=self._array_library)

    def _concatenate_coefficients(self, ham_coefficients, dis_coefficients):
        if self._hamiltonian_operators is not None and self._dissipator_operators is not None:
            return _numpy_multi_dispatch(ham_coefficients, dis_coefficients, path="append", axis=-1)
        if self._hamiltonian_operators is not None and self._dissipator_operators is None:
            return ham_coefficients
        if self._hamiltonian_operators is None and self._dissipator_operators is not None:
            return dis_coefficients

        return None


class ScipySparseVectorizedLindbladCollection(VectorizedLindbladCollection):
    r"""Scipy sparse version of VectorizedLindbladCollection."""

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        decimals: Optional[int] = 10,
    ):
        r"""Initialize collection.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            static_dissipators: Dissipator terms with coefficient 1.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
            decimals: Decimals to round the sparse operators to 0.
        """
        self._decimals = decimals
        super().__init__(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
        )

    def _convert_to_array_type(self, obj: any) -> csr_matrix:
        if obj is None:
            return None

        return numpy_alias(like="scipy_sparse").asarray(np.round(obj, self._decimals))

    def _convert_to_array_type_list(self, obj: any) -> np.ndarray[csr_matrix]:
        return _to_csr_object_array(obj, self._decimals)

    def _construct_operator_collection(self, *args, **kwargs):
        """The class used for evaluating the vectorized model or RHS."""
        return ScipySparseOperatorCollection(*args, **kwargs)


def package_density_matrices(y: ArrayLike) -> ArrayLike:
    """Sends an array ``y`` of density matrices to a ``(1,)`` array of dtype object, where entry
    ``[0]`` is ``y``. Formally avoids for-loops through vectorization.

    Args:
        y: An array.
    Returns:
        Array with dtype object.
    """
    # As written here, only works for (n,n) Arrays
    obj_arr = np.empty(shape=(1,), dtype="O")
    obj_arr[0] = y
    return obj_arr


# Using vectorization with signature, works on (k,n,n) Arrays -> (k,1) Array
package_density_matrices = np.vectorize(package_density_matrices, signature="(n,n)->(1)")


def unpackage_density_matrices(y: ArrayLike) -> ArrayLike:
    """Inverse function of :func:`package_density_matrices`.

    Since this function is much slower than packaging, avoid it unless absolutely needed (as in case
    of passing multiple density matrices to :meth:`SparseLindbladCollection.evaluate_rhs`).

    Args:
        y: An array to extract the first element from.
    Returns:
        A ``(k,n,n)`` array.
    """
    return y[0]


unpackage_density_matrices = np.vectorize(unpackage_density_matrices, signature="(1)->(n,n)")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Registering linear_combo functions to alias. This computes a linear combination of matrices (given
by a 3d array).
"""

import numpy as np


def register_linear_combo(alias):
    """Register linear functions for each array library."""

    @alias.register_default(path="linear_combo")
    def _(coeffs, mats):
        return np.tensordot(coeffs, mats, axes=1)

    @alias.register_function(lib="numpy", path="linear_combo")
    def _(coeffs, mats):
        return np.tensordot(coeffs, mats, axes=1)

    try:
        import jax.numpy as jnp

        @alias.register_function(lib="jax", path="linear_combo")
        def _(coeffs, mats):
            # real and imag broken up to avoid real/complex tensordot warning
            return jnp.tensordot(coeffs, mats.real, axes=1) + 1j * jnp.tensordot(
                coeffs, mats.imag, axes=1
            )

        from jax.experimental.sparse import sparsify

        jsparse_sum = sparsify(jnp.sum)

        @alias.register_function(lib="jax_sparse", path="linear_combo")
        def _(coeffs, mats):
            # pylint: disable=unexpected-keyword-arg
            return jsparse_sum(jnp.broadcast_to(coeffs[:, None, None], mats.shape) * mats, axis=0)

    except ImportError:
        pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""This module implements the job class used for DynamicsBackend objects."""

from typing import Callable, Dict
from datetime import datetime

from qiskit.providers.backend import Backend
from qiskit.providers import JobV1 as Job
from qiskit.providers import JobStatus, JobError


class DynamicsJob(Job):
    """DynamicsJob class for DynamicsBackend."""

    _async = False

    def __init__(
        self, backend: Backend, job_id: str, fn: Callable, fn_kwargs: Dict, **kwargs
    ) -> None:
        """Initializes the job.

        Args:
            backend: The backend used to run the job.
            job_id: A unique id in the context of the backend used to run the job.
            fn: Function to run the simulation.
            fn_kwargs: Kwargs for fn.
            kwargs: Any key value metadata to associate with this job.
        """
        super().__init__(backend, job_id, **kwargs)
        self._fn = fn
        self._fn_kwargs = fn_kwargs
        self._result = None
        self._time_per_step = {"CREATED": datetime.now()}

    def submit(self):
        """Run the simulation.

        Raises:
            JobError: if trying to re-submit the job.
        """
        if self._result is not None:
            raise JobError("Dynamics job has already been submitted.")
        self._result = self._fn(job_id=self.job_id(), **self._fn_kwargs)
        self._time_per_step["COMPLETED"] = datetime.now()

    def result(self):
        """Get job result.

        Returns:
            qiskit.Result: Result object.

        Raises:
            JobError: If job has not been submitted.
        """
        if self._result is None:
            raise JobError("Job has not been submitted.")
        return self._result

    def status(self):
        """Gets the status of the job.

        Returns:
            JobStatus: The current JobStatus.
        """
        if self._result is None:
            return JobStatus.INITIALIZING

        return JobStatus.DONE

    def time_per_step(self) -> Dict:
        """Return the date and time information on each step of the job processing.

        Returns:
            Dict for time of creation and time of completion of job.
        """
        return self._time_per_step



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
=========================================================
Perturbation Theory (:mod:`qiskit_dynamics.perturbation`)
=========================================================

.. currentmodule:: qiskit_dynamics.perturbation

This module contains tools for numerically computing and utilizing perturbation theory terms.
Perturbation theory is an advanced topic; a brief review of the concepts and notation required to
understand the contents in this module are given in the :ref:`Time-dependent perturbation theory and
multi-variable series expansions review <perturbation review>` discussion.

.. _td perturbation theory:

Time-dependent perturbation theory
==================================

The function :func:`.solve_lmde_perturbation` computes Dyson series :footcite:`dyson_radiation_1949`
and Magnus expansion :footcite:`magnus_exponential_1954,blanes_magnus_2009` terms in a
multi-variable setting via algorithms in :footcite:`puzzuoli_algorithms_2023`. It can also be used
to compute Dyson-like integrals using the algorithm in :footcite:`haas_engineering_2019`. Results
are returned in either a :class:`PowerSeriesData` or :class:`DysonLikeData` class, which are data
classes with functionality for indexing and accessing specific perturbation terms. See the function
documentation for further details.

Truncated power-series representation and multisets
===================================================

The class :class:`.ArrayPolynomial` represents an array-valued multivariable polynomial (i.e. a
truncated power series), and provides functionality for both evaluating and transforming
array-valued polynomials.

This module makes use of the `multiset package <https://pypi.org/project/multiset/>`_ for indexing
multi-variable power series. See the :ref:`multiset and power series notation section <multiset
power series>` of the perturbation review for an explanation of this convention.


Perturbation module functions
=============================

.. autosummary::
    :toctree: ../stubs/

    solve_lmde_perturbation


Perturbation module classes
===========================

.. autosummary::
    :toctree: ../stubs/

    ArrayPolynomial
    PowerSeriesData
    DysonLikeData

.. footbibliography::
"""

from .array_polynomial import ArrayPolynomial
from .solve_lmde_perturbation import solve_lmde_perturbation
from .perturbation_data import PowerSeriesData, DysonLikeData



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""Utilities model module."""

from typing import Union, List
import numpy as np
from scipy.sparse import csr_matrix, issparse
from scipy.sparse import kron as sparse_kron
from scipy.sparse import identity as sparse_identity

from qiskit_dynamics.arraylias.alias import ArrayLike, _to_dense, _numpy_multi_dispatch


def _kron(A, B):
    return _numpy_multi_dispatch(A, B, path="kron")


def vec_commutator(
    A: Union[ArrayLike, csr_matrix, List[csr_matrix]],
) -> Union[ArrayLike, csr_matrix, List[csr_matrix]]:
    r"""Linear algebraic vectorization of the linear map X -> -i[A, X]
    in column-stacking convention. In column-stacking convention we have

    .. math::
        vec(ABC) = C^T \otimes A vec(B),

    so for the commutator we have

    .. math::
        -i[A, \cdot] = -i(A \cdot - \cdot A \mapsto id \otimes A - A^T \otimes id)

    Note: this function is also "vectorized" in the programming sense for dense arrays.

    Args:
        A: Either a 2d array representing the matrix A described above,
           a 3d array representing a list of matrices, a sparse matrix, or a
           list of sparse matrices.

    Returns:
        ArrayLike: Vectorized version of the map.
    """

    if issparse(A):
        # single, sparse matrix
        sp_iden = sparse_identity(A.shape[-1], format="csr")
        return -1j * (sparse_kron(sp_iden, A) - sparse_kron(A.T, sp_iden))
    if isinstance(A, (list, np.ndarray)) and issparse(A[0]):
        # taken to be 1d array of 2d sparse matrices
        sp_iden = sparse_identity(A[0].shape[-1], format="csr")
        out = [-1j * (sparse_kron(sp_iden, mat) - sparse_kron(mat.T, sp_iden)) for mat in A]
        return np.array(out)

    A = _to_dense(A)
    iden = np.eye(A.shape[-1])
    axes = list(range(A.ndim))
    axes[-1] = axes[-2]
    axes[-2] += 1
    return -1j * (_kron(iden, A) - _kron(A.transpose(axes), iden))


def vec_dissipator(
    L: Union[ArrayLike, csr_matrix, List[csr_matrix]],
) -> Union[ArrayLike, csr_matrix, List[csr_matrix]]:
    r"""Linear algebraic vectorization of the linear map
    X -> L X L^\dagger - 0.5 * (L^\dagger L X + X L^\dagger L)
    in column stacking convention.

    This gives

    .. math::
        \overline{L} \otimes L - 0.5(id \otimes L^\dagger L +
            (L^\dagger L)^T \otimes id)

    Note: this function is also "vectorized" in the programming sense for dense matrices.
    """

    if issparse(L):
        sp_iden = sparse_identity(L[0].shape[-1], format="csr")
        return sparse_kron(L.conj(), L) - 0.5 * (
            sparse_kron(sp_iden, L.conj().T * L) + sparse_kron(L.T * L.conj(), sp_iden)
        )
    if isinstance(L, (list, np.ndarray)) and issparse(L[0]):
        # taken to be 1d array of 2d sparse matrices
        sp_iden = sparse_identity(L[0].shape[-1], format="csr")
        out = [
            sparse_kron(mat.conj(), mat)
            - 0.5
            * (sparse_kron(sp_iden, mat.conj().T * mat) + sparse_kron(mat.T * mat.conj(), sp_iden))
            for mat in L
        ]
        return np.array(out)

    iden = np.eye(L.shape[-1])
    axes = list(range(L.ndim))

    L = _to_dense(L)
    axes[-1] = axes[-2]
    axes[-2] += 1
    Lconj = L.conj()
    LdagL = Lconj.transpose(axes) @ L
    LdagLtrans = LdagL.transpose(axes)

    return _kron(Lconj, iden) @ _kron(iden, L) - 0.5 * (
        _kron(iden, LdagL) + _kron(LdagLtrans, iden)
    )



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
Register rmatmul functions to alias.
"""

import numpy as np


def register_rmatmul(alias):
    """Register rmatmul functions to each array library."""

    @alias.register_function(lib="numpy", path="rmatmul")
    def _(x, y):
        return np.matmul(y, x)

    @alias.register_function(lib="scipy_sparse", path="rmatmul")
    def _(x, y):
        return y * x

    try:
        from jax.experimental import sparse as jsparse
        import jax.numpy as jnp

        jsparse_matmul = jsparse.sparsify(jnp.matmul)

        @alias.register_function(lib="jax", path="rmatmul")
        def _(x, y):
            return jnp.matmul(y, x)

        @alias.register_function(lib="jax_sparse", path="rmatmul")
        def _(x, y):
            return jsparse_matmul(y, x)

    except ImportError:
        pass



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Registering asarray functions to alias
"""

from typing import Iterable

import numpy as np
from scipy.sparse import csr_matrix, issparse


def _isinstance_qutip_qobj(obj):
    """Check if the object is a qutip Qobj.
    Args:
        obj (any): Any object for testing.
    Returns:
        Bool: True if obj is qutip Qobj
    """
    return type(obj).__name__ == "Qobj"


def register_asarray(alias):
    """register asarray functions to each array libraries"""

    @alias.register_default(path="asarray")
    def _(arr):
        if _isinstance_qutip_qobj(arr):
            return arr.full()
        return np.asarray(arr)

    @alias.register_function(lib="scipy_sparse", path="asarray")
    def _(arr):
        if issparse(arr) or (isinstance(arr, Iterable) and issparse(arr[0])):
            return arr
        return csr_matrix(arr)

    try:
        from jax.experimental.sparse import BCOO

        @alias.register_function(lib="jax_sparse", path="asarray")
        def _(arr):
            if type(arr).__name__ == "BCOO":
                return arr
            return BCOO.fromdense(arr)

    except ImportError:
        pass



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
============================================
Arraylias (:mod:`qiskit_dynamics.arraylias`)
============================================

.. currentmodule:: qiskit_dynamics.arraylias

This module contains Qiskit Dynamics-global extensions of the default NumPy and SciPy aliases
provided by `Arraylias <https://qiskit-community.github.io/arraylias/>`_. These are used to manage
dispatching of array operations for the different array types supported by Qiskit Dynamics. They
have been configured beyond the `Arraylias <https://qiskit-community.github.io/arraylias/>`_
defaults to additionally support both `JAX <https://jax.readthedocs.io/en/latest/>`_ and SciPy
sparse types. The following table summarizes the registered libraries and respective types.

.. list-table:: Supported libraries
    :widths: 10 70
    :header-rows: 1

    * - ``array_library``
      - Registered types
    * - ``"numpy"``
      - Default supported by the Arraylias NumPy and SciPy aliases.
    * - ``"jax"``
      - Default supported by the Arraylias NumPy and SciPy aliases.
    * - ``"jax_sparse"``
      - The JAX ``jax.experimental.sparse.BCOO`` array type.
    * - ``"scipy_sparse"``
      - Subclasses of the ``scipy.sparse.spmatrix`` sparse base class. When instantiating SciPy
        sparse arrays, the alias will specifically create ``scipy.sparse.csr_matrix`` instances.

The global configured aliases and the aliased libraries can be imported from
``qiskit_dynamics.arraylias``, and are summarized by the following table.

.. list-table:: Configured Arraylias objects
    :widths: 10 70
    :header-rows: 1

    * - Arraylias object
      - Description
    * - ``DYNAMICS_NUMPY_ALIAS``
      - Qiskit Dynamics-global NumPy alias.
    * - ``DYNAMICS_SCIPY_ALIAS``
      - Qiskit Dynamics-global SciPy alias.
    * - ``DYNAMICS_NUMPY``
      - Qiskit Dynamics-global aliased NumPy library.
    * - ``DYNAMICS_SCIPY``
      - Qiskit Dynamics-global aliased SciPy library.
"""

from .alias import (
    DYNAMICS_NUMPY_ALIAS,
    DYNAMICS_SCIPY_ALIAS,
    DYNAMICS_NUMPY,
    DYNAMICS_SCIPY,
    ArrayLike,
    requires_array_library,
)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
====================================
Pulse (:mod:`qiskit_dynamics.pulse`)
====================================

.. currentmodule:: qiskit_dynamics.pulse

This module contains tools to interface :mod:`qiskit.pulse` with Qiskit Dynamics. Qiskit Dynamics
simulates time evolution using the :class:`Signal` class, however :mod:`qiskit.pulse` specifies
pulse instructions using a :class:`~qiskit.pulse.Schedule` or :class:`~qiskit.pulse.ScheduleBlock`.
This module contains the required converters to convert from a :mod:`qiskit.pulse` control
specification into :class:`Signal` instances for simulation.

Converters
==========

The conversion from a :class:`~qiskit.pulse.Schedule` to a list of :class:`Signal` instances is done
with the :class:`InstructionToSignals` converter. The following codeblock shows a simple example
instantiation, and how to use it to convert a :class:`~qiskit.pulse.Schedule` to a list of
:class:`Signal` instances.

.. warning::

    The code blocks below supress ``DeprecationWarning`` instances raised by Qiskit Pulse in
    `qiskit` `1.3`.

.. jupyter-execute::
    :hide-code:

    # silence deprecation warnings from pulse
    import warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning)

.. code-block:: python

    converter = InstructionToSignals(dt=1, carriers=None)
    signals = converter.get_signals(sched)

An example schedule, and the corresponding converted signals, is shown below.

.. jupyter-execute::
    :hide-code:

    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec

    import qiskit.pulse as pulse

    from qiskit_dynamics.pulse import InstructionToSignals


    with pulse.build(name="schedule") as sched:
        pulse.play(pulse.Drag(20, 0.5, 4, 0.5), pulse.DriveChannel(0))
        pulse.shift_phase(1.0, pulse.DriveChannel(0))
        pulse.play(pulse.Drag(20, 0.5, 4, 0.5), pulse.DriveChannel(0))
        pulse.shift_frequency(0.5, pulse.DriveChannel(0))
        pulse.play(pulse.GaussianSquare(200, 0.3, 4, 150), pulse.DriveChannel(0))
        pulse.play(pulse.GaussianSquare(200, 0.3, 4, 150), pulse.DriveChannel(1))

    fig = plt.figure(constrained_layout=True, figsize=(10, 7))
    spec = gridspec.GridSpec(ncols=2, nrows=2, figure=fig)
    ax1 = fig.add_subplot(spec[0, :])
    ax2 = fig.add_subplot(spec[1, 0])
    ax3 = fig.add_subplot(spec[1, 1])

    converter = InstructionToSignals(dt=1, carriers=None)

    signals = converter.get_signals(sched)

    signals[0].draw(0, 239, 400, axis=ax2, title="Signal from DriveChannel(0)")
    signals[1].draw(0, 239, 400, axis=ax3, title="Signal from DriveChannel(1)")
    sched.draw(axis=ax1)

Converter class
===============

.. autosummary::
   :toctree: ../stubs/

   InstructionToSignals
"""
from .pulse_to_signals import InstructionToSignals



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Register multiply functions to alias.
"""


def register_multiply(alias):
    """Register multiply functions to each array library."""

    @alias.register_fallback(path="multiply")
    def _(x, y):
        return x * y

    @alias.register_function(lib="scipy_sparse", path="multiply")
    def _(x, y):
        return x.multiply(y)

    try:
        from jax.experimental import sparse as jsparse
        import jax.numpy as jnp

        jsparse_multiply = jsparse.sparsify(jnp.multiply)

        @alias.register_function(lib="jax_sparse", path="multiply")
        def _(x, y):
            return jsparse_multiply(x, y)

    except ImportError:
        pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Module for custom binary operations between arrays.
"""

from typing import Optional, List, Tuple, Callable

import numpy as np

from qiskit_dynamics.arraylias.alias import _preferred_lib

try:
    import jax.numpy as jnp
    from jax import vmap
except ImportError:
    pass


class _CustomBinaryOp:
    r"""A binary operation between arrays of dimension >1d built from taking linear combinations of
    a base binary operation acting on sub-arrays.

    This class constructs customized binary operations between lists of arrays :math:`A = (A_i)` and
    :math:`B = (B_i)` of the form:

    .. math::
        (A \times B)_i = \sum_{jk} a_{ijk} f(A_j, B_k),

    where :math:`a_{ijk}` is an array of complex scalars, and :math:`f` is a binary operation (a
    common example being matrix multiplication).

    At instantiation the binary operation :math:`f`, as well as the array :math:`a_{ijk}`, are
    specified. The array :math:`a_{ijk}` is given in a sparse format, as a list where each entry
    specifies the 2d sub-array :math:`a_{i}` as a 2-tuple with entries:

        - The non-zero entries of :math:`a_i` given as a list.
        - A 2d-array with each entry being the pair of indices ``[j,k]``.

    Internally, this specification is translated into one more suited for efficient evaluation:
        - A specification of the unique pairs of required evaluations :math:`f(A_j, B_k)` in terms
          of two arrays giving the left indices and the right indices.
        - A 2-tuple of 2d arrays specifying the linear combination of unique evaluations of
          :math:`f` to compute the :math:`i^{th}` entry of the binary operations. The :math:`i^{th}`
          entry of each array is:
            - The list of coefficients in the linear combination.
            - The index of the unique product for each coefficient.
          These arrays are padded with the value ``-1`` to make each linear combo the same length
          (relevant for JAX evaluation).
    """

    def __init__(
        self,
        operation_rule: List,
        binary_op: Callable,
        index_offset: Optional[int] = 0,
        operation_rule_compiled: Optional[bool] = False,
    ):
        """Initialize the binary operation.

        Args:
            operation_rule: Rule for the binary op as described in the doc string.
            binary_op: The binary operation.
            index_offset: Shift to be added to the indices in operation_rule.
            operation_rule_compiled: True if the operation_rule already corresponds to a rule that
                has been compiled to the internal representation.
        """

        # store binary op and compile rule to internal format for evaluation
        self._binary_op = binary_op

        if not operation_rule_compiled:
            operation_rule = _compile_custom_operation_rule(operation_rule, index_offset)
        self._unique_evaluation_pairs, self._linear_combo_rule = operation_rule

    def __call__(self, A: np.ndarray, B: np.ndarray) -> np.ndarray:
        """Evaluate the custom binary operation on arrays A, B."""

        lib = _preferred_lib(A, B)

        if lib == "jax":
            unique_evaluations = _compute_unique_evaluations_jax(
                A, B, self._unique_evaluation_pairs, vmap(self._binary_op)
            )
            return _compute_linear_combos_jax(unique_evaluations, self._linear_combo_rule)

        unique_evaluations = _compute_unique_evaluations(
            A, B, self._unique_evaluation_pairs, self._binary_op
        )
        return _compute_linear_combos(unique_evaluations, self._linear_combo_rule)


class _CustomMatmul(_CustomBinaryOp):
    """Custom matmul multiplication."""

    def __init__(
        self,
        operation_rule: List,
        index_offset: Optional[int] = 0,
        operation_rule_compiled: Optional[bool] = False,
    ):
        """Initialize."""

        super().__init__(
            operation_rule=operation_rule,
            binary_op=lambda A, B: A @ B,
            index_offset=index_offset,
            operation_rule_compiled=operation_rule_compiled,
        )


class _CustomMul(_CustomBinaryOp):
    """Custom mul multiplication."""

    def __init__(
        self,
        operation_rule: List,
        index_offset: Optional[int] = 0,
        operation_rule_compiled: Optional[bool] = False,
    ):
        """Initialize."""

        super().__init__(
            operation_rule=operation_rule,
            binary_op=lambda A, B: A * B,
            index_offset=index_offset,
            operation_rule_compiled=operation_rule_compiled,
        )


def _compile_custom_operation_rule(
    operation_rule: List,
    index_offset: Optional[int] = 0,
    unique_evaluation_len: Optional[int] = None,
    linear_combo_len: Optional[int] = None,
) -> Tuple[np.array, np.array]:
    """Compile the list of unique evaluations and linear combinations required to implement a given
    operation_rule.

    See _CustomBinaryOp doc string for formatting details.

    Args:
        operation_rule: Custom operation rule in the sparse format in the _CustomBinaryOp doc
            string.
        index_offset: Integer specifying a shift to apply in the 2nd and 3rd indices in the sparse
            representation of :math:`a_{ijk}`.
        unique_evaluation_len: Integer specifying a minimum length to represent the unique
            multiplications list. The unique multiplication list is padded with entries ``[-1, -1]``
            to meet the minimum length.
        linear_combo_len: Minimum length for linear combo specification. Coefficients are padded
            with zeros, and the unique multiplication indices are padded with ``-1``.

    Returns:
        Tuple[np.array, np.array]: Multiplication rule compiled into a list of unique products and a
            list of linear combinations of unique products for implementing the custom dot rule.
    """

    # force numpy usage for algebra specification
    new_rule = []
    for coeffs, index_pairs in operation_rule:
        new_rule.append((np.array(coeffs), np.array(index_pairs, dtype=int) + index_offset))

    operation_rule = tuple(new_rule)

    # construct list of unique multiplications and linear combo rule
    unique_evaluation_list = []
    linear_combo_rule = []
    for coeffs, index_pairs in operation_rule:
        sub_linear_combo = []
        for index_pair in index_pairs:
            # convert to list to avoid array comparison issues
            index_pair = list(index_pair)
            if index_pair not in unique_evaluation_list:
                unique_evaluation_list.append(index_pair)

            sub_linear_combo.append(unique_evaluation_list.index(index_pair))

        linear_combo_rule.append((coeffs, np.array(sub_linear_combo, dtype=int)))

    unique_evaluation_pairs = np.array(unique_evaluation_list, dtype=int)

    if unique_evaluation_len is not None and unique_evaluation_len > len(unique_evaluation_pairs):
        padding = -1 * np.ones((unique_evaluation_len - len(unique_evaluation_pairs), 2), dtype=int)
        unique_evaluation_pairs = np.append(unique_evaluation_pairs, padding, axis=0)

    # pad linear combo rule with -1 for shorter rules
    max_len = linear_combo_len or 0
    for coeffs, _ in linear_combo_rule:
        max_len = max(max_len, len(coeffs))

    padded_linear_combo_rule = []
    for coeffs, indices in linear_combo_rule:
        if coeffs.shape[0] < max_len:
            pad_len = max_len - coeffs.shape[0]
            coeffs = np.append(coeffs, np.zeros(pad_len))
            indices = np.append(indices, -1 * np.ones(pad_len, dtype=int))

        padded_linear_combo_rule.append((coeffs, indices))

    coeff_array = np.vstack([a[0] for a in padded_linear_combo_rule])
    index_array = np.vstack([a[1] for a in padded_linear_combo_rule])

    linear_combo_rule = (coeff_array, index_array)

    return unique_evaluation_pairs, linear_combo_rule


def _compute_unique_evaluations(
    A: np.array, B: np.array, unique_evaluation_pairs: np.array, binary_op: Callable
) -> np.array:
    """Compute ``binary_op(A[j], B[k])`` for index pairs ``[j, k]`` in
    ``unique_evaluation_pairs``.
    """

    # evaluate first pair (assumes not all evaluation pairs are paddings of [-1, -1])
    eval_pair = unique_evaluation_pairs[0]
    unique_evaluation = binary_op(A[eval_pair[0]], B[eval_pair[1]])

    M0 = np.zeros(unique_evaluation.shape, dtype=complex)
    unique_evaluations = np.empty(
        (len(unique_evaluation_pairs),) + unique_evaluation.shape, dtype=complex
    )
    unique_evaluations[0] = unique_evaluation

    for idx, eval_pair in enumerate(unique_evaluation_pairs[1:]):
        if eval_pair[0] == -1:
            unique_evaluations[idx + 1] = M0
        else:
            unique_evaluations[idx + 1] = binary_op(A[eval_pair[0]], B[eval_pair[1]])

    return unique_evaluations


def _compute_linear_combos(
    unique_evaluations: np.array, linear_combo_rule: Tuple[np.array, np.array]
) -> np.array:
    r"""Compute linear combinations of the entries in the array ``unique_mults`` according to
    ``linear_combo_rule``. The :math:`j^{th}` entry of the output is given by
    :math:`sum_k c[j, k] unique_mults[idx[j, k]]`, where ``linear_combo_rule`` is ``(c, idx)``.
    """
    M0 = np.zeros_like(unique_evaluations[0])
    C = np.empty((len(linear_combo_rule[0]),) + unique_evaluations[0].shape, dtype=complex)
    for idx, (coeffs, eval_indices) in enumerate(zip(linear_combo_rule[0], linear_combo_rule[1])):
        mat = M0
        for coeff, eval_idx in zip(coeffs, eval_indices):
            if idx != -1:
                mat = mat + coeff * unique_evaluations[eval_idx]
        C[idx] = mat

    return C


def _compute_unique_evaluations_jax(
    A: np.ndarray,
    B: np.ndarray,
    unique_evaluation_pairs: np.array,
    binary_op: Callable,
) -> np.ndarray:
    """JAX version of a single loop step of :meth:`linear_combos`. Note that in this function
    binary_op is assumed to be vectorized."""
    A = jnp.append(A, jnp.zeros((1,) + A[0].shape, dtype=complex), axis=0)
    B = jnp.append(B, jnp.zeros((1,) + B[0].shape, dtype=complex), axis=0)

    return binary_op(A[unique_evaluation_pairs[:, 0]], B[unique_evaluation_pairs[:, 1]])


def _compute_single_linear_combo_jax(
    unique_evaluations: np.ndarray, single_combo_rule: Tuple[np.array, np.array]
) -> np.ndarray:
    """JAX version of :meth:`unique_products`."""
    coeffs, indices = single_combo_rule
    return jnp.tensordot(coeffs, unique_evaluations[indices], axes=1)


try:
    _compute_linear_combos_jax = vmap(_compute_single_linear_combo_jax, in_axes=(None, (0, 0)))
except NameError:
    pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Base classes and abstract subsystem operators.
"""

from abc import ABC, abstractmethod
from typing import Optional, Union, List
from copy import copy

import numpy as np

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import ArrayLike

from .subsystem import Subsystem


class AbstractSubsystemOperator(ABC):
    """Abstract base class for operators acting on Subsystems."""

    def __init__(self, subsystems: Union[Subsystem, List[Subsystem]]):
        """Initialize with the list of subsystems the operator acts on."""
        if isinstance(subsystems, Subsystem):
            subsystems = [subsystems]
        self._subsystems = subsystems

    @abstractmethod
    def base_matrix(self):
        """Return the matrix defined on the internal subsystems."""

    @abstractmethod
    def __str__(self):
        """String representation."""

    def __repr__(self):
        return str(self)

    @property
    def subsystems(self) -> List[Subsystem]:
        """Get the subsystems the operator acts on."""
        return self._subsystems

    def matrix(self, ordered_subsystems: Optional[List[Subsystem]] = None):
        """Build the matrix for the operator relative to the ordered subsystems."""

        if ordered_subsystems is None:
            ordered_subsystems = self.subsystems

        if any(subsystem not in ordered_subsystems for subsystem in self.subsystems):
            raise QiskitError("Attempted to build matrix but missing subsystem.")

        return _matrix_implicit_identity(self.base_matrix(), self.subsystems, ordered_subsystems)

    def restrict_subsystems(self, subsystems: List[Subsystem]) -> "AbstractSubsystemOperator":
        """Reduce the operator to the list of subsystems.

        Components of operators with support on removed subsystems will be set to 0.
        """
        if any(subsystem not in subsystems for subsystem in self.subsystems):
            new_subsystems = []
            for subsystem in subsystems:
                if subsystem in self.subsystems:
                    new_subsystems.append(subsystem)

            return ZeroOperator(new_subsystems)

        return self

    def remove_subsystems(self, subsystems: List[Subsystem]) -> "AbstractSubsystemOperator":
        """Return operator with subsystems removed.

        Components of operators with support on removed subsystems will be set to 0.
        """
        if any(x in subsystems for x in self.subsystems):
            new_subsystems = []
            for subsystem in self.subsystems:
                if subsystem not in subsystems:
                    new_subsystems.append(subsystem)

            return ZeroOperator(new_subsystems)

        return self

    def __add__(self, other):
        """Add other to self."""

        if _isscalar(other):
            other = ScalarOperator(other, subsystems=self.subsystems)

        if isinstance(other, ZeroOperator):
            return self

        if isinstance(self, ZeroOperator):
            return other

        return OperatorSum(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __matmul__(self, other):
        """Multiply other with self."""

        if _isscalar(other):
            raise QiskitError("matmul of a scalar with a AbstractSubsystemOperator is not defined.")

        if isinstance(other, ZeroOperator):
            return other

        if isinstance(self, ZeroOperator):
            return self

        return OperatorMatmul(self, other)

    def __rmatmul__(self, other):
        """Could add more validation here."""
        if _isscalar(other):
            raise QiskitError("matmul of a scalar with a AbstractSubsystemOperator is not defined.")

    def __mul__(self, other):
        if isinstance(other, ZeroOperator):
            return other

        if isinstance(self, ZeroOperator):
            return self

        if _isscalar(other):
            return ScalarOperatorProduct(other, self)

        return OperatorMul(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)


class ScalarOperator(AbstractSubsystemOperator):
    """A scalar operator."""

    def __init__(self, value, subsystems):
        """Initialize.

        Args:
            value: The value of the scalar.
            subsystems: List of subsystems.
        """
        self._value = value
        super().__init__(subsystems)

    @property
    def value(self):
        """The scalar value."""
        return self._value

    def base_matrix(self):
        total_dim = int(np.prod([x.dim for x in self.subsystems]))
        return self.value * np.eye(total_dim, dtype=complex)

    def __str__(self):
        return str(self.value)


class ZeroOperator(ScalarOperator):
    """The zero operator."""

    def __init__(self, subsystems):
        """Initialize.

        Args:
            subsystems: List of subsystems.
        """
        super().__init__(value=0.0, subsystems=subsystems)

    def base_matrix(self):
        if len(self.subsystems) == 0:
            raise QiskitError("Attempting to create matrix of zero operator with no subsystems.")
        total_dim = int(np.prod([x.dim for x in self.subsystems]))
        return np.zeros((total_dim, total_dim), dtype=complex)


class CompositeOperator(AbstractSubsystemOperator):
    """An operator built from multiple sub operators."""

    def __init__(self, operators):
        """Initialize.

        Args:
            operators: A list of operators.
        """
        subsystems = copy(operators[0].subsystems)
        for op in operators[1:]:
            for x in op.subsystems:
                if x not in subsystems:
                    subsystems.append(x)

        self._operators = operators
        super().__init__(subsystems)

    @property
    def operators(self):
        """The operators making up the composite operator."""
        return self._operators


class OperatorSum(CompositeOperator):
    """Sum of operators A + B."""

    def __init__(self, a, b):
        """Initialize.

        Args:
            a: An operator.
            b: An operator.
        """
        super().__init__(operators=[a, b])

    def base_matrix(self):
        return self.operators[0].matrix(self.subsystems) + self.operators[1].matrix(self.subsystems)

    def restrict_subsystems(self, subsystems: List[Subsystem]) -> "SubsystemOperator":
        """Reduce the operator to the list of subsystems.

        Components of operators with support on removed subsystems will be set to 0.
        """
        return self.operators[0].restrict_subsystems(subsystems) + self.operators[
            1
        ].restrict_subsystems(subsystems)

    def remove_subsystems(self, subsystems: List[Subsystem]) -> "SubsystemOperator":
        """Return operator with subsystems removed.

        Components of operators with support on removed subsystems will be set to 0.
        """
        return self.operators[0].remove_subsystems(subsystems) + self.operators[
            1
        ].remove_subsystems(subsystems)

    def __str__(self):
        if isinstance(self.operators[0], CompositeOperator):
            op1_string = f"({self.operators[0]})"
        else:
            op1_string = str(self.operators[0])

        if isinstance(self.operators[1], CompositeOperator):
            op2_string = f"({self.operators[1]})"
        else:
            op2_string = str(self.operators[1])
        return f"{op1_string} + {op2_string}"


class OperatorMatmul(CompositeOperator):
    """Matmul of operators A @ B."""

    def __init__(self, a, b):
        super().__init__(operators=[a, b])

    def base_matrix(self):
        return self.operators[0].matrix(self.subsystems) @ self.operators[1].matrix(self.subsystems)

    def __str__(self):
        if isinstance(self.operators[0], CompositeOperator):
            op1_string = f"({self.operators[0]})"
        else:
            op1_string = str(self.operators[0])

        if isinstance(self.operators[1], CompositeOperator):
            op2_string = f"({self.operators[1]})"
        else:
            op2_string = str(self.operators[1])
        return f"{op1_string} @ {op2_string}"


class OperatorMul(CompositeOperator):
    """Mul of operators A * B."""

    def __init__(self, a, b):
        super().__init__(operators=[a, b])

    def base_matrix(self):
        return self.operators[0].matrix(self.subsystems) * self.operators[1].matrix(self.subsystems)

    def __str__(self):
        if isinstance(self.operators[0], CompositeOperator):
            op1_string = f"({self.operators[0]})"
        else:
            op1_string = str(self.operators[0])

        if isinstance(self.operators[1], CompositeOperator):
            op2_string = f"({self.operators[1]})"
        else:
            op2_string = str(self.operators[1])
        return f"{op1_string} * {op2_string}"


class ScalarOperatorProduct(AbstractSubsystemOperator):
    """Product of a scalar and an operator."""

    def __init__(self, scalar, operator):
        """Initialize.

        Args:
            scalar: The scalar.
            operator: The operator.
        """
        self._scalar = scalar
        self._operator = operator
        super().__init__(operator.subsystems)

    @property
    def scalar(self):
        """The scalar."""
        return self._scalar

    def base_matrix(self):
        return self._scalar * self._operator.matrix()

    def __str__(self):
        return f"{self.scalar} * {self._operator}"


class FunctionOperator(AbstractSubsystemOperator):
    """A function applied on an operator. This assumes the output is the same shape/dimension as the
    input.
    """

    def __init__(self, func, operator, func_name=None):
        """Initialize.

        Args:
            func: The function.
            operator: The operator.
            func_name: The name of the function.
        """
        self._func = func
        self._operator = operator
        self._func_name = func_name
        super().__init__(operator.subsystems)

    @property
    def func(self):
        """The function applied to the operator."""
        return self._func

    def base_matrix(self):
        return self.func(self._operator.matrix())

    def __str__(self):
        func_name = self._func_name or "f"
        return f"{func_name}({self._operator})"


def _matrix_implicit_identity(matrix, matrix_subsystems, target_subsystems):
    """Given a matrix defined on a tensor product system described by matrix_subsystems, return the
    definition of the matrix as it acts on target_subsystems in the given order, with implicit
    identity on the subsystems in the complement of matrix_subsystems.
    """
    if any(s not in target_subsystems for s in matrix_subsystems):
        raise QiskitError("matrix_subsystems not all contained in target_subsystems.")

    subsystems_complement = []
    for s in target_subsystems:
        if s not in matrix_subsystems:
            subsystems_complement.append(s)

    if len(subsystems_complement) > 0:
        total_dim = np.prod([s.dim for s in subsystems_complement])
        matrix = unp.kron(matrix, np.eye(total_dim, dtype=matrix.dtype))
        matrix_subsystems = subsystems_complement + matrix_subsystems

    # from now on assume matrix_subsystems has the same elements as target_subsystems
    subsystem_dims = [s.dim for s in matrix_subsystems]
    subsystem_dims.reverse()

    matrix = matrix.reshape(subsystem_dims + subsystem_dims)

    num_subsystems = len(matrix_subsystems)
    matrix_subsystems = list(reversed(matrix_subsystems))
    target_subsystems = list(reversed(target_subsystems))

    reorder = [matrix_subsystems.index(x) for x in target_subsystems]
    reorder = reorder + [x + num_subsystems for x in reorder]

    matrix = matrix.transpose(reorder)
    full_dim = np.prod(subsystem_dims)
    return matrix.reshape((full_dim, full_dim))


def _isscalar(x):
    """Check if x is a scalar."""
    if isinstance(x, ArrayLike):
        try:
            x = unp.asarray(x)
            if x.ndim == 0:
                return True
        # pylint: disable=broad-exception-caught
        except Exception:
            pass
    return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Lindblad model module.
"""

from typing import Tuple, Union, List, Optional
from warnings import warn

from qiskit import QiskitError
from qiskit_dynamics.arraylias.alias import ArrayLike, _to_dense, _to_dense_list
from qiskit_dynamics.signals import Signal, SignalList
from .generator_model import (
    BaseGeneratorModel,
    _static_operator_into_frame_basis,
    _operators_into_frame_basis,
)
from .hamiltonian_model import HamiltonianModel, is_hermitian
from .operator_collections import (
    LindbladCollection,
    ScipySparseLindbladCollection,
    VectorizedLindbladCollection,
    ScipySparseVectorizedLindbladCollection,
)
from .rotating_frame import RotatingFrame

try:
    import jax
except ImportError:
    pass


class LindbladModel(BaseGeneratorModel):
    r"""A model of a quantum system in terms of the Lindblad master equation.

    The Lindblad master equation models the evolution of a density matrix according to:

    .. math::

        \dot{\rho}(t) = -i[H(t), \rho(t)] + \mathcal{D}_0(\rho(t)) + \mathcal{D}(t)(\rho(t)),

    where :math:`\mathcal{D}_0` is the static dissipator portion, and :math:`\mathcal{D}(t)` is the
    time-dependent dissipator portion of the equation, given by

    .. math::

        \mathcal{D}_0(\rho(t)) = \sum_j N_j \rho N_j^\dagger
                                      - \frac{1}{2} \{N_j^\dagger N_j, \rho\},

    and

    .. math::

        \mathcal{D}(t)(\rho(t)) = \sum_j \gamma_j(t) L_j \rho L_j^\dagger
                                  - \frac{1}{2} \{L_j^\dagger L_j, \rho\},

    respectively. In the above:

        - :math:`[\cdot, \cdot]` and :math:`\{\cdot, \cdot\}` respectively denote the matrix
          commutator and anti-commutator,
        - :math:`H(t)` denotes the Hamiltonian,
        - :math:`N_j` denotes the operators appearing in the static dissipator,
        - :math:`L_j` denotes the operators appearing in the time-dpendent portion of the
          dissipator, and
        - :math:`\gamma_j(t)` denotes the signal corresponding to the
          :math:`j^{th}` time-dependent dissipator operator.

    Instantiating an instance of :class:`~qiskit_dynamics.models.LindbladModel` requires specifying
    the above decomposition:

    .. code-block:: python

        lindblad_model = LindbladModel(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            hamiltonian_signals=hamiltonian_signals,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
            dissipator_signals=dissipator_signals
        )

    where the arguments ``hamiltonian_operators``, ``hamiltonian_signals``, and
    ``static_hamiltonian`` are for the Hamiltonian decomposition as in
    :class:`~qiskit_dynamics.models.HamiltonianModel`,
    and the ``static_dissipators`` correspond to the :math:`N_j`, the ``dissipator_operators``
    to the :math:`L_j`, and the ``dissipator_signals`` the :math:`\gamma_j(t)`.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        hamiltonian_signals: Optional[Union[List[Signal], SignalList]] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        dissipator_signals: Optional[Union[List[Signal], SignalList]] = None,
        rotating_frame: Optional[Union[ArrayLike, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        array_library: Optional[str] = None,
        vectorized: bool = False,
        validate: bool = True,
    ):
        """Initialize.

        Args:
            static_hamiltonian: Constant term in Hamiltonian.
            hamiltonian_operators: List of operators in Hamiltonian with time-dependent
                coefficients.
            hamiltonian_signals: Time-dependent coefficients for hamiltonian_operators.
            static_dissipators: List of dissipators with coefficient 1.
            dissipator_operators: List of dissipator operators with time-dependent coefficients.
            dissipator_signals: Time-dependent coefficients for dissipator_operators.
            rotating_frame: Rotating frame in which calcualtions are to be done. If provided, it is
                assumed that all operators were already in the frame basis.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                frame operator is diagonalized.
            array_library: Array library with which to represent the operators in the model, and to
                evaluate the model. See the list of supported array libraries in the
                :mod:`.arraylias` submodule API documentation. If ``None``, the arrays will be
                handled by general dispatching rules.
            vectorized: Whether or not to setup the Lindblad equation in vectorized mode.
                If ``True``, the operators in the model are stored as :math:`(dim^2,dim^2)` matrices
                that act on vectorized density matrices by left-multiplication. Setting this to
                ``True`` is necessary for ``SuperOp`` simulation.
            validate: If True check input hamiltonian_operators and static_hamiltonian are
                Hermitian.

        Raises:
            QiskitError: If model insufficiently or incorrectly specified.
        """

        if (
            static_hamiltonian is None
            and hamiltonian_operators is None
            and static_dissipators is None
            and dissipator_operators is None
        ):
            raise QiskitError(
                f"{type(self).__name__} requires at least one of static_hamiltonian "
                "hamiltonian_operators, static_dissipators, or dissipator_operators "
                "to be specified at construction."
            )

        if validate:
            if (static_hamiltonian is not None) and (not is_hermitian(static_hamiltonian)):
                raise QiskitError("""LinbladModel static_hamiltonian must be Hermitian.""")
            if (hamiltonian_operators is not None) and any(
                not is_hermitian(op) for op in hamiltonian_operators
            ):
                raise QiskitError("""LindbladModel hamiltonian_operators must be Hermitian.""")

        self._vectorized = vectorized
        self._rotating_frame = RotatingFrame(rotating_frame)
        self._in_frame_basis = in_frame_basis

        # jax sparse arrays cannot be used directly at this stage
        setup_library = array_library
        if array_library == "jax_sparse":
            setup_library = "jax"

        # set up internal operators
        if static_hamiltonian is not None:
            static_hamiltonian = -1j * static_hamiltonian
        static_hamiltonian = _static_operator_into_frame_basis(
            static_operator=static_hamiltonian,
            rotating_frame=self._rotating_frame,
            array_library=setup_library,
        )
        if static_hamiltonian is not None:
            static_hamiltonian = 1j * static_hamiltonian

        hamiltonian_operators = _operators_into_frame_basis(
            operators=hamiltonian_operators,
            rotating_frame=self._rotating_frame,
            array_library=setup_library,
        )

        static_dissipators = _operators_into_frame_basis(
            operators=static_dissipators,
            rotating_frame=self._rotating_frame,
            array_library=setup_library,
        )

        dissipator_operators = _operators_into_frame_basis(
            operators=dissipator_operators,
            rotating_frame=self._rotating_frame,
            array_library=setup_library,
        )

        self._operator_collection = _get_lindblad_operator_collection(
            array_library=array_library,
            vectorized=vectorized,
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
        )

        self.signals = (hamiltonian_signals, dissipator_signals)

        super().__init__(array_library=array_library)

    @classmethod
    def from_hamiltonian(
        cls,
        hamiltonian: HamiltonianModel,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        dissipator_signals: Optional[ArrayLike] = None,
        array_library: Optional[str] = None,
        vectorized: bool = False,
    ):
        """Construct from a :class:`HamiltonianModel`.

        Args:
            hamiltonian: The :class:`HamiltonianModel`.
            static_dissipators: List of dissipators with coefficient 1.
            dissipator_operators: List of dissipators with time-dependent coefficients.
            dissipator_signals: List time-dependent coefficients for dissipator_operators.
            array_library: Array library to use.
            vectorized: Whether or not to vectorize the Lindblad equation.

        Returns:
            LindbladModel: Linblad model from parameters.
        """

        # store whether hamiltonian is in_frame_basis and set to False
        in_frame_basis = hamiltonian.in_frame_basis
        hamiltonian.in_frame_basis = False

        # get the operators
        static_hamiltonian = hamiltonian.static_operator
        hamiltonian_operators = hamiltonian.operators

        # return to previous value
        hamiltonian.in_frame_basis = in_frame_basis

        return cls(
            static_hamiltonian=_to_dense(static_hamiltonian),
            hamiltonian_operators=_to_dense_list(hamiltonian_operators),
            hamiltonian_signals=hamiltonian.signals,
            static_dissipators=_to_dense_list(static_dissipators),
            dissipator_operators=_to_dense_list(dissipator_operators),
            dissipator_signals=dissipator_signals,
            rotating_frame=hamiltonian.rotating_frame,
            in_frame_basis=hamiltonian.in_frame_basis,
            array_library=array_library,
            vectorized=vectorized,
        )

    @property
    def dim(self) -> int:
        """The matrix dimension."""
        if self._operator_collection.static_hamiltonian is not None:
            return self._operator_collection.static_hamiltonian.shape[-1]
        elif self._operator_collection.hamiltonian_operators is not None:
            return self._operator_collection.hamiltonian_operators[0].shape[-1]
        elif self._operator_collection.static_dissipators is not None:
            return self._operator_collection.static_dissipators[0].shape[-1]
        else:
            return self._operator_collection.dissipator_operators[0].shape[-1]

    @property
    def signals(self) -> Tuple[SignalList]:
        """The model's signals as tuple with the 0th entry storing the Hamiltonian signals and the
        1st entry storing the dissipator signals.

        Raises:
            QiskitError: If, when setting this property, the given signals are incompatible with
                operator structure.
        """
        return (self._hamiltonian_signals, self._dissipator_signals)

    @signals.setter
    def signals(self, new_signals: Tuple[List[Signal]]):
        hamiltonian_signals, dissipator_signals = new_signals

        # set Hamiltonian signals
        if hamiltonian_signals is None:
            self._hamiltonian_signals = None
        elif hamiltonian_signals is not None and self.hamiltonian_operators is None:
            raise QiskitError("Hamiltonian signals must be None if hamiltonian_operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(hamiltonian_signals, list):
                hamiltonian_signals = SignalList(hamiltonian_signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(hamiltonian_signals, SignalList):
                raise QiskitError("Hamiltonian signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.hamiltonian_operators, list):
                len_hamiltonian_operators = len(self.hamiltonian_operators)
            else:
                len_hamiltonian_operators = self.hamiltonian_operators.shape[0]
            if len(hamiltonian_signals) != len_hamiltonian_operators:
                raise QiskitError(
                    "Hamiltonian signals need to have the same length as Hamiltonian operators."
                )

            self._hamiltonian_signals = hamiltonian_signals

        # set dissipator signals
        if dissipator_signals is None:
            self._dissipator_signals = None
        elif dissipator_signals is not None and self.dissipator_operators is None:
            raise QiskitError("Dissipator signals must be None if dissipator_operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(dissipator_signals, list):
                dissipator_signals = SignalList(dissipator_signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(dissipator_signals, SignalList):
                raise QiskitError("Dissipator signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.dissipator_operators, list):
                len_dissipator_operators = len(self.dissipator_operators)
            else:
                len_dissipator_operators = self.dissipator_operators.shape[0]
            if len(dissipator_signals) != len_dissipator_operators:
                raise QiskitError(
                    "Dissipator signals need to have the same length as dissipator operators."
                )

            self._dissipator_signals = dissipator_signals

    @property
    def in_frame_basis(self) -> bool:
        """Whether to represent the model in the basis in which the frame operator is
        diagonalized.
        """
        return self._in_frame_basis

    @in_frame_basis.setter
    def in_frame_basis(self, in_frame_basis: bool):
        self._in_frame_basis = in_frame_basis

    @property
    def static_hamiltonian(self) -> ArrayLike:
        """The static Hamiltonian term."""
        if self._operator_collection.static_hamiltonian is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.static_hamiltonian
        return self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.static_hamiltonian
        )

    @property
    def hamiltonian_operators(self) -> ArrayLike:
        """The Hamiltonian operators."""
        if self._operator_collection.hamiltonian_operators is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.hamiltonian_operators
        return self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.hamiltonian_operators
        )

    @property
    def static_dissipators(self) -> ArrayLike:
        """The static dissipator operators."""
        if self._operator_collection.static_dissipators is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.static_dissipators
        return self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.static_dissipators
        )

    @property
    def dissipator_operators(self) -> ArrayLike:
        """The dissipator operators."""
        if self._operator_collection.dissipator_operators is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.dissipator_operators
        return self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.dissipator_operators
        )

    @property
    def vectorized(self) -> bool:
        """Whether or not the Lindblad equation is vectorized."""
        return self._vectorized

    @property
    def rotating_frame(self):
        """The rotating frame."""
        return self._rotating_frame

    def evaluate_hamiltonian(self, time: float) -> ArrayLike:
        """Evaluates Hamiltonian matrix at a given time.

        Args:
            time: The time at which to evaluate the Hamiltonian.

        Returns:
            ArrayLike: Hamiltonian matrix.
        """

        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)

        ham = self._operator_collection.evaluate_hamiltonian(hamiltonian_sig_vals)
        if self.rotating_frame.frame_diag is not None:
            ham = self.rotating_frame.operator_into_frame(
                time,
                ham,
                operator_in_frame_basis=True,
                return_in_frame_basis=self._in_frame_basis,
                vectorized_operators=self.vectorized,
            )

        return ham

    def evaluate(self, time: float) -> ArrayLike:
        """Evaluate the model in array format as a matrix, independent of state.

        Args:
            time: The time to evaluate the model at.

        Returns:
            ArrayLike: The evaluated model as an anti-Hermitian matrix.

        Raises:
            QiskitError: If model cannot be evaluated.
            NotImplementedError: If the model is currently unvectorized.
        """
        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)
        elif self._operator_collection.hamiltonian_operators is not None:
            raise QiskitError(
                f"{type(self).__name__} with non-empty hamiltonian operators cannot be evaluated "
                "without hamiltonian signals."
            )

        dissipator_sig_vals = None
        if self._dissipator_signals is not None:
            dissipator_sig_vals = self._dissipator_signals(time)
        elif self._operator_collection.dissipator_operators is not None:
            raise QiskitError(
                f"{type(self).__name__} with non-empty dissipator operators cannot be evaluated "
                "without dissipator signals."
            )

        if self.vectorized:
            out = self._operator_collection.evaluate(hamiltonian_sig_vals, dissipator_sig_vals)
            return self.rotating_frame.vectorized_map_into_frame(
                time, out, operator_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
            )
        else:
            raise NotImplementedError(
                "Non-vectorized Lindblad models cannot be represented without a given state."
            )

    def evaluate_rhs(self, time: float, y: ArrayLike) -> ArrayLike:
        """Evaluates the Lindblad model at a given time.

        Args:
            time: The time at which the model should be evaluated.
            y: Density matrix as an (n,n) Array if not vectorized, or an (n^2) Array if using
                vectorized evaluation.

        Returns:
            ArrayLike: Either the evaluated generator or the state.

        Raises:
            QiskitError: If signals not sufficiently specified.
        """

        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)
        elif self._operator_collection.hamiltonian_operators is not None:
            raise QiskitError(
                f"{type(self).__name__} with non-empty hamiltonian operators cannot be evaluated "
                "without hamiltonian signals."
            )

        dissipator_sig_vals = None
        if self._dissipator_signals is not None:
            dissipator_sig_vals = self._dissipator_signals(time)
        elif self._operator_collection.dissipator_operators is not None:
            raise QiskitError(
                f"{type(self).__name__} with non-empty dissipator operators cannot be evaluated "
                "without dissipator signals."
            )

        if self.rotating_frame.frame_diag is not None:
            # Take y out of the frame, but keep in the frame basis
            rhs = self.rotating_frame.operator_out_of_frame(
                time,
                y,
                operator_in_frame_basis=self._in_frame_basis,
                return_in_frame_basis=True,
                vectorized_operators=self.vectorized,
            )

            rhs = self._operator_collection.evaluate_rhs(
                hamiltonian_sig_vals, dissipator_sig_vals, rhs
            )

            # Put rhs back into the frame, potentially converting its basis.
            rhs = self.rotating_frame.operator_into_frame(
                time,
                rhs,
                operator_in_frame_basis=True,
                return_in_frame_basis=self._in_frame_basis,
                vectorized_operators=self.vectorized,
            )

        else:
            rhs = self._operator_collection.evaluate_rhs(
                hamiltonian_sig_vals, dissipator_sig_vals, y
            )

        return rhs


def _get_lindblad_operator_collection(
    array_library: Optional[str],
    vectorized: bool,
    static_hamiltonian: Optional[ArrayLike],
    hamiltonian_operators: Optional[ArrayLike],
    static_dissipators: Optional[ArrayLike],
    dissipator_operators: Optional[ArrayLike],
) -> Union[
    LindbladCollection,
    ScipySparseLindbladCollection,
    VectorizedLindbladCollection,
    ScipySparseVectorizedLindbladCollection,
]:
    """Construct a Lindblad operator collection.

    Args:
        array_library: Array library to use.
        vectorized: Whether or not to vectorize the Lindblad equation.
        static_hamiltonian: Constant part of the Hamiltonian.
        hamiltonian_operators: Operators in Hamiltonian with time-dependent coefficients.
        static_dissipators: Dissipation operators with coefficient 1.
        dissipator_operators: Dissipation operators with variable coefficients.

    Returns:
        Union[
            LindbladCollection,
            ScipySparseLindbladCollection,
            VectorizedLindbladCollection,
            ScipySparseVectorizedLindbladCollection,
        ]: Right-hand side evaluation object.
    """

    operator_kwargs = {
        "static_hamiltonian": static_hamiltonian,
        "hamiltonian_operators": hamiltonian_operators,
        "static_dissipators": static_dissipators,
        "dissipator_operators": dissipator_operators,
    }

    if array_library == "scipy_sparse":
        if vectorized:
            return ScipySparseVectorizedLindbladCollection(**operator_kwargs)

        return ScipySparseLindbladCollection(**operator_kwargs)

    if array_library == "jax_sparse":
        # warn that sparse mode when using JAX is primarily recommended for use on CPU
        if jax.default_backend() != "cpu":
            warn(
                """Using sparse mode with JAX is primarily recommended for use on CPU.""",
                stacklevel=2,
            )

    if vectorized:
        return VectorizedLindbladCollection(**operator_kwargs, array_library=array_library)

    return LindbladCollection(**operator_kwargs, array_library=array_library)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Custom fixed step solvers.
"""

from typing import Callable, Optional, Tuple
from warnings import warn
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult
from scipy.linalg import expm

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias import ArrayLike, requires_array_library

try:
    import jax
    from jax import vmap
    import jax.numpy as jnp
    from jax.lax import scan, cond, associative_scan
    from jax.scipy.linalg import expm as jexpm
except ImportError:
    pass

from .solver_utils import merge_t_args, trim_t_results
from .lanczos import lanczos_expm
from .lanczos import jax_lanczos_expm


def RK4_solver(
    rhs: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """Fixed step RK4 solver.

    Args:
        rhs: Callable, either a generator rhs
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """
    div6 = 1.0 / 6

    def take_step(rhs_func, t, y, h):
        h2 = 0.5 * h
        t_plus_h2 = t + h2

        k1 = rhs_func(t, y)
        k2 = rhs_func(t_plus_h2, y + h2 * k1)
        k3 = rhs_func(t_plus_h2, y + h2 * k2)
        k4 = rhs_func(t + h, y + h * k3)

        return y + div6 * h * (k1 + 2 * k2 + 2 * k3 + k4)

    return fixed_step_solver_template(
        take_step, rhs_func=rhs, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


def scipy_expm_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
    magnus_order: int = 1,
):
    """Fixed-step size matrix exponential based solver implemented with
    ``scipy.linalg.expm``. Solves the specified problem by taking steps of
    size no larger than ``max_dt``.

    Args:
        generator: Generator for the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.
        magnus_order: The expansion order in the Magnus method. Only orders 1, 2 and 3
            are supported.

    Returns:
        OdeResult: Results object.
    """
    take_step = get_exponential_take_step(magnus_order, expm_func=expm)

    return fixed_step_solver_template(
        take_step, rhs_func=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


def lanczos_diag_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    k_dim: int,
    t_eval: Optional[ArrayLike] = None,
):
    """Fixed-step size matrix exponential based solver implemented using
    lanczos algorithm. Solves the specified problem by taking steps of
    size no larger than ``max_dt``.

    Args:
        generator: Generator for the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        k_dim: Integer which specifies the dimension of Krylov subspace used for
            lanczos iteration. Acts as an accuracy parameter. ``k_dim`` must
            always be less than or equal to dimension of generator.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """

    def take_step(generator, t0, y, h):
        eval_time = t0 + (h / 2)
        return lanczos_expm(generator(eval_time), y, k_dim, h)

    return fixed_step_solver_template(
        take_step, rhs_func=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


@requires_array_library("jax")
def jax_lanczos_diag_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    k_dim: int,
    t_eval: Optional[ArrayLike] = None,
):
    """JAX version of lanczos_diag_solver."""

    def take_step(generator, t0, y, h):
        eval_time = t0 + (h / 2)
        return jax_lanczos_expm(generator(eval_time), y, k_dim, h)

    y0 = unp.asarray(y0, dtype=complex)

    return fixed_step_solver_template_jax(
        take_step, rhs_func=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


@requires_array_library("jax")
def jax_RK4_solver(
    rhs: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """JAX version of RK4_solver.

    Args:
        rhs: Callable, either a generator rhs
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """
    div6 = 1.0 / 6

    def take_step(rhs_func, t, y, h):
        h2 = 0.5 * h
        t_plus_h2 = t + h2

        k1 = rhs_func(t, y)
        k2 = rhs_func(t_plus_h2, y + h2 * k1)
        k3 = rhs_func(t_plus_h2, y + h2 * k2)
        k4 = rhs_func(t + h, y + h * k3)

        return y + div6 * h * (k1 + 2 * k2 + 2 * k3 + k4)

    return fixed_step_solver_template_jax(
        take_step, rhs_func=rhs, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


@requires_array_library("jax")
def jax_RK4_parallel_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """Parallel version of :func:`jax_RK4_solver` specialized to LMDEs.

    Args:
        generator: Generator of the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """
    dim = y0.shape[-1]
    ident = jnp.eye(dim, dtype=complex)

    div6 = 1.0 / 6

    def take_step(generator, t, h):
        h2 = 0.5 * h
        gh2 = generator(t + h2)

        k1 = generator(t)
        k2 = gh2 @ (ident + h2 * k1)
        k3 = gh2 @ (ident + h2 * k2)
        k4 = generator(t + h) @ (ident + h * k3)

        return ident + div6 * h * (k1 + 2 * k2 + 2 * k3 + k4)

    return fixed_step_lmde_solver_parallel_template_jax(
        take_step, generator=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


@requires_array_library("jax")
def jax_expm_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
    magnus_order: int = 1,
):
    """Fixed-step size matrix exponential based solver implemented with ``jax``.
    Solves the specified problem by taking steps of size no larger than ``max_dt``.

    Args:
        generator: Generator for the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.
        magnus_order: The expansion order in the Magnus method. Only orders 1, 2, and 3
            are supported.

    Returns:
        OdeResult: Results object.
    """
    take_step = get_exponential_take_step(magnus_order, expm_func=jexpm)

    return fixed_step_solver_template_jax(
        take_step, rhs_func=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


@requires_array_library("jax")
def jax_expm_parallel_solver(
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
    magnus_order: int = 1,
):
    """Parallel version of :func:`jax_expm_solver` implemented with JAX parallel operations.

    Args:
        generator: Generator for the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.
        magnus_order: The expansion order in the Magnus method. Only orders 1, 2, and 3
            are supported.

    Returns:
        OdeResult: Results object.
    """
    take_step = get_exponential_take_step(magnus_order, expm_func=jexpm, just_propagator=True)

    return fixed_step_lmde_solver_parallel_template_jax(
        take_step, generator=generator, t_span=t_span, y0=y0, max_dt=max_dt, t_eval=t_eval
    )


def matrix_commutator(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    """Compute the commutator of two matrices.

    Args:
        m1: First argument to the commutator.
        m2: Second argument to the commutator.

    Returns:
        Matrix commutator of ``m1`` and ``m2``.
    """
    return m1 @ m2 - m2 @ m1


def get_exponential_take_step(
    magnus_order: int, expm_func: Callable, just_propagator: bool = False
):
    """Return a function implementing the infinitessimal magnus solver at 1st, 2nd, and 3rd
    Magnus orders, specified by the user. See also the documentation of :func:`scipy_expm_solver`
    for details.

    Args:
        magnus_order: The expansion order in the Magnus method. Only accepts values in
            ``[1, 2, 3]``.
        expm_func: Method of matrix exponentian.
        just_propagator: Whether or not to return function that only computes propagator.
            If False, returns a function with signature f(generator, t0, y, h), and if True, returns
            a function with signature f(generator, t0, h).

    Returns:
        take_step: Infinitessimal exponential Magnus solver.

    Raises:
        QiskitError: If ``magnus_order`` not in ``[1, 2, 3]``.
    """
    # if clause based on magnus order
    if magnus_order == 1:

        def propagator(generator, t0, h):
            return expm_func(generator(t0 + (h / 2)) * h)

    elif magnus_order == 2:
        # second-order step size constants
        c1 = 0.5 - np.sqrt(3) / 6
        c2 = 0.5 + np.sqrt(3) / 6
        p2 = np.sqrt(3) / 12

        def propagator(generator, t0, h):
            # midpoint generator calls
            g1 = generator(t0 + c1 * h)
            g2 = generator(t0 + c2 * h)

            # Magnus terms
            terms = h * (g1 + g2) / 2 + p2 * (h**2) * matrix_commutator(g2, g1)

            # solution
            return expm_func(terms)

    elif magnus_order == 3:
        # third-order step size constants
        d1 = 0.5 - np.sqrt(15) / 10
        d2 = 0.5
        d3 = 0.5 + np.sqrt(15) / 10
        c0 = np.sqrt(15) / 3
        c1 = 10.0 / 3

        def propagator(generator, t0, h):
            # midpoint generator calls
            g1 = generator(t0 + d1 * h)
            g2 = generator(t0 + d2 * h)
            g3 = generator(t0 + d3 * h)

            # linear combinations of generators
            a1 = h * g2
            a2 = c0 * h * (g3 - g1)
            a3 = c1 * h * (g3 - 2 * g2 + g1)

            # intermediate commutators
            comm1 = matrix_commutator(a1, a2)
            comm2 = matrix_commutator(2 * a3 + comm1, a1) / 60

            # Magnus terms
            terms = a1 + (a3 / 12) + matrix_commutator(-20 * a1 - a3 + comm1, a2 + comm2) / 240

            # solution
            return expm_func(terms)

    else:
        raise QiskitError("Only magnus_order 1, 2, and 3 are supported.")

    if just_propagator:
        return propagator

    def take_step(generator, t0, y, h):
        return propagator(generator, t0, h) @ y

    return take_step


def fixed_step_solver_template(
    take_step: Callable,
    rhs_func: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """Helper function for implementing fixed-step solvers supporting both
    ``t_span`` and ``max_dt`` arguments. ``take_step`` is assumed to be a
    function implementing a single step of size h of a fixed-step method.
    The signature of ``take_step`` is assumed to be:
        - rhs_func: Either a generator :math:`G(t)` or RHS function :math:`f(t,y)`.
        - t0: The current time.
        - y0: The current state.
        - h: The size of the step to take.

    It returns:
        - y: The state of the DE at time t0 + h.

    ``take_step`` is used to integrate the DE specified by ``rhs_func``
    through all points in ``t_eval``, taking steps no larger than ``max_dt``.
    Each interval in ``t_eval`` is divided into the least number of sub-intervals
    of equal length so that the sub-intervals are smaller than ``max_dt``.

    Args:
        take_step: Callable for fixed step integration.
        rhs_func: Callable, either a generator or rhs function.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """

    y0 = unp.asarray(y0)

    t_list, h_list, n_steps_list = get_fixed_step_sizes(t_span, t_eval, max_dt)

    ys = [y0]
    for current_t, h, n_steps in zip(t_list, h_list, n_steps_list):
        y = ys[-1]
        inner_t = current_t
        for _ in range(n_steps):
            y = take_step(rhs_func, inner_t, y, h)
            inner_t = inner_t + h
        ys.append(y)
    ys = unp.asarray(ys)

    results = OdeResult(t=t_list, y=ys)

    return trim_t_results(results, t_eval)


def fixed_step_solver_template_jax(
    take_step: Callable,
    rhs_func: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """This function is the jax control-flow version of
    :meth:`fixed_step_solver_template`. See the documentation of :meth:`fixed_step_solver_template`
    for details.

    Args:
        take_step: Callable for fixed step integration.
        rhs_func: Callable, either a generator or rhs function.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """

    y0 = jnp.array(y0)

    t_list, h_list, n_steps_list = get_fixed_step_sizes(t_span, t_eval, max_dt)

    # if jax, need bound on number of iterations in each interval
    max_steps = n_steps_list.max()

    def identity(y):
        return y

    # interval integrator set up for jax.lax.scan
    def scan_interval_integrate(carry, x):
        current_t, h, n_steps = x
        current_y = carry

        def scan_take_step(carry, step):
            t, y = carry
            y = cond(step < n_steps, lambda y: take_step(rhs_func, t, y, h), identity, y)
            t = t + h
            return (t, y), None

        next_y = scan(scan_take_step, (current_t, current_y), jnp.arange(max_steps))[0][1]

        return next_y, next_y

    ys = scan(
        scan_interval_integrate,
        init=y0,
        xs=(jnp.array(t_list[:-1]), jnp.array(h_list), jnp.array(n_steps_list)),
    )[1]

    ys = jnp.append(jnp.expand_dims(y0, axis=0), ys, axis=0)

    results = OdeResult(t=t_list, y=ys)

    return trim_t_results(results, t_eval)


def fixed_step_lmde_solver_parallel_template_jax(
    take_step: Callable,
    generator: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    max_dt: float,
    t_eval: Optional[ArrayLike] = None,
):
    """Parallelized and LMDE specific version of fixed_step_solver_template_jax.

    Assuming the structure of an LMDE:
    * Computes all propagators over each individual time-step in parallel using ``jax.vmap``.
    * Computes all propagators from t_span[0] to each intermediate time point in parallel
      using ``jax.lax.associative_scan``.
    * Applies results to y0 and extracts the desired time points from ``t_eval``.

    The above logic is slightly varied to save some operations is ``y0`` is square.

    The signature of ``take_step`` is assumed to be:
        - generator: A generator :math:`G(t)`.
        - t: The current time.
        - h: The size of the step to take.

    It returns:
        - y: The state of the DE at time t + h.

    Note that this differs slightly from the other template functions, in that
    ``take_step`` does not take take in ``y``, the state at time ``t``. The
    parallelization procedure described above uses the initial state being the identity
    matrix for each time step, and thus it is unnecessary to supply this to ``take_step``.

    Args:
        take_step: Fixed step integration rule.
        generator: Generator for the LMDE.
        t_span: Interval to solve over.
        y0: Initial state.
        max_dt: Maximum step size.
        t_eval: Optional list of time points at which to return the solution.

    Returns:
        OdeResult: Results object.
    """

    # warn the user that the parallel solver will be very slow if run on a cpu
    if jax.default_backend() == "cpu":
        warn(
            """JAX parallel solvers will likely run slower on CPUs than non-parallel solvers.
            To make use of their capabilities it is recommended to use a GPU.""",
            stacklevel=2,
        )

    y0 = jnp.array(y0)

    t_list, h_list, n_steps_list = get_fixed_step_sizes(t_span, t_eval, max_dt)

    # set up time information for computing propagators in parallel
    all_times = []  # all stepping points
    all_h = []  # step sizes for each point above
    t_list_locations = [0]  # ordered list of locations in all_times that are in t_list
    for t, h, n_steps in zip(t_list, h_list, n_steps_list):
        all_times = np.append(all_times, t + h * np.arange(n_steps))
        all_h = np.append(all_h, h * np.ones(n_steps))
        t_list_locations = np.append(t_list_locations, [t_list_locations[-1] + n_steps])

    # compute propagators over each time step in parallel
    step_propagators = vmap(lambda t, h: take_step(generator, t, h))(all_times, all_h)

    # multiply propagators together in parallel
    ys = None

    def reverse_mul(A, B):
        return jnp.matmul(B, A)

    if y0.ndim == 2 and y0.shape[0] == y0.shape[1]:
        # if square, append y0 as the first step propagator, scan, and extract
        intermediate_props = associative_scan(
            reverse_mul, jnp.append(jnp.array([y0]), step_propagators, axis=0), axis=0
        )
        ys = intermediate_props[t_list_locations]
    else:
        # if not square, scan propagators, extract relevant time points, multiply by y0,
        # then prepend y0
        intermediate_props = associative_scan(reverse_mul, step_propagators, axis=0)
        # intermediate_props doesn't include t0, so shift t_list_locations when extracting
        intermediate_y = intermediate_props[t_list_locations[1:] - 1] @ y0
        ys = jnp.append(jnp.array([y0]), intermediate_y, axis=0)

    results = OdeResult(t=t_list, y=ys)

    return trim_t_results(results, t_eval)


def get_fixed_step_sizes(
    t_span: ArrayLike, t_eval: ArrayLike, max_dt: float
) -> Tuple[ArrayLike, ArrayLike, ArrayLike]:
    """Merge ``t_span`` and ``t_eval``, and determine the number of time steps and
    and step sizes (no larger than ``max_dt``) required to fixed-step integrate between
    each time point.

    Args:
        t_span: Total interval of integration.
        t_eval: Time points within t_span at which the solution should be returned.
        max_dt: Max size step to take.

    Returns:
        Tuple[Array, Array, Array]: with merged time point list, list of step sizes to take
        between time points, and list of corresponding number of steps to take between time steps.
    """
    # time args are non-differentiable
    t_span = np.array(t_span)
    max_dt = np.array(max_dt)
    t_list = np.array(merge_t_args(t_span, t_eval))

    # set the number of time steps required in each interval so that
    # no steps larger than max_dt are taken
    delta_t_list = np.diff(t_list)
    n_steps_list = np.abs(delta_t_list / max_dt).astype(int)

    # correct potential rounding errors
    for idx, (delta_t, n_steps) in enumerate(zip(delta_t_list, n_steps_list)):
        if n_steps == 0:
            n_steps_list[idx] = 1
        # absolute value to handle backwards integration
        elif np.abs(delta_t / n_steps) / max_dt > 1 + 1e-15:
            n_steps_list[idx] = n_steps + 1

    # step size in each interval
    h_list = np.array(delta_t_list / n_steps_list)

    return t_list, h_list, n_steps_list
